from typing import Union, Any

from nebullvm.core.models import ModelParams, QuantizationType
from nebullvm.operations.optimizations.compilers.base import Compiler

from nebullvm.optional_modules.torch import (
    torch,
    Module,
    GraphModule,
)
from nebullvm.tools.data import DataManager
from nebullvm.tools.transformations import MultiStageTransformation


class TorchDynamoCompiler(Compiler):
    supported_ops = {
        "cpu": [None],
        "gpu": [None],
    }

    def execute(
        self,
        model: Module,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        metric_drop_ths: float = None,
        quantization_type: QuantizationType = None,
        input_data: DataManager = None,
        **kwargs,
    ):
        """Optimize the input model using pytorch built-in techniques.

        Args:
            model (torch.nn.Module): The pytorch model.
            model_params (ModelParams): The model parameters.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction. Default: None.
            metric_drop_ths (float, optional): Threshold for the accepted drop
                in terms of precision. Any optimized model with a higher drop
                will be ignored. Default: None.
            quantization_type (QuantizationType, optional): The desired
                quantization algorithm to be used. Default: None.
            input_data (DataManager): User defined data. Default: None.
        """

        if quantization_type not in self.supported_ops[self.device.type.value]:
            self.compiled_model = None
            return

        self.logger.info(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )

        self.compiled_model = self._compile_model(model, model_params)

    @torch.no_grad()
    def _compile_model(
        self,
        model: Union[Module, GraphModule],
        network_parameters: ModelParams,
    ) -> Any:
        dynamic = False
        if network_parameters.dynamic_info is not None:
            dynamic = True
        return torch.compile(model, dynamic=dynamic)

    def _quantize_model(self, **kwargs) -> Any:
        raise NotImplementedError



import platform

from nebullvm.core.models import (
    DeepLearningFramework,
    DeviceType,
    ModelCompiler,
)
from nebullvm.operations.optimizations.optimizers.base import Optimizer
from nebullvm.operations.optimizations.compilers.utils import (
    tvm_is_available,
    bladedisc_is_available,
    deepsparse_is_available,
    intel_neural_compressor_is_available,
    torch_tensorrt_is_available,
    onnxruntime_is_available,
    tensorrt_is_available,
    openvino_is_available,
    torch_neuron_is_available,
    torch_xla_is_available,
    faster_transformer_is_available,
)
from nebullvm.optional_modules.torch import torch
from nebullvm.optional_modules.utils import (
    torch_is_available,
    tensorflow_is_available,
    onnx_is_available,
)
from nebullvm.tools.utils import check_module_version


class PytorchOptimizer(Optimizer):
    def __init__(self):
        super().__init__()
        self.pipeline_dl_framework = DeepLearningFramework.PYTORCH

    def _select_compilers_from_hardware(self):
        compilers = []
        if torch_is_available():
            if self.device.type is DeviceType.TPU:
                if torch_xla_is_available():
                    compilers.append(ModelCompiler.TORCH_XLA)
                else:
                    raise RuntimeError(
                        "Torch XLA is not available on your platform. "
                        "Please install torch-xla the readme at this "
                        "link: https://github.com/pytorch/xla"
                    )
            elif self.device.type is DeviceType.NEURON:
                if torch_neuron_is_available():
                    compilers.append(ModelCompiler.TORCH_NEURON)
                else:
                    raise RuntimeError(
                        "Torch Neuron is not available on your platform. "
                        "Please install torch-neuron by following "
                        "this guide: https://awsdocs-neuron"
                        ".readthedocs-hosted.com/en/latest/general/"
                        "quick-start/torch-neuron.html."
                    )
            else:
                compilers.append(ModelCompiler.TORCHSCRIPT)
                if (
                    check_module_version(torch, min_version="2.0.0")
                    and platform.system() != "Windows"
                    and False
                ):  # Deactivated because save and load methods are
                    # not implemented
                    compilers.append(ModelCompiler.TORCH_DYNAMO)
                if tvm_is_available():
                    compilers.append(ModelCompiler.APACHE_TVM_TORCH)
                if bladedisc_is_available():
                    compilers.append(ModelCompiler.BLADEDISC)

                if self.device.type is DeviceType.CPU:
                    if deepsparse_is_available():
                        compilers.append(ModelCompiler.DEEPSPARSE)
                    if intel_neural_compressor_is_available():
                        compilers.append(ModelCompiler.INTEL_NEURAL_COMPRESSOR)
                elif self.device.type is DeviceType.GPU:
                    if torch_tensorrt_is_available():
                        compilers.append(ModelCompiler.TENSOR_RT_TORCH)
                    if faster_transformer_is_available():
                        compilers.append(ModelCompiler.FASTER_TRANSFORMER)
        return compilers


class TensorflowOptimizer(Optimizer):
    def __init__(self):
        super().__init__()
        self.pipeline_dl_framework = DeepLearningFramework.TENSORFLOW

    def _select_compilers_from_hardware(self):
        compilers = []
        if tensorflow_is_available():
            compilers.append(ModelCompiler.XLA)
            compilers.append(ModelCompiler.TFLITE)
        return compilers


class ONNXOptimizer(Optimizer):
    def __init__(self):
        super().__init__()
        self.pipeline_dl_framework = DeepLearningFramework.NUMPY

    def _select_compilers_from_hardware(self):
        compilers = []
        if onnx_is_available():
            if onnxruntime_is_available():
                compilers.append(ModelCompiler.ONNX_RUNTIME)
            if tvm_is_available():
                compilers.append(ModelCompiler.APACHE_TVM_ONNX)
            if self.device.type is DeviceType.GPU and tensorrt_is_available():
                compilers.append(ModelCompiler.TENSOR_RT_ONNX)
            if self.device.type is DeviceType.CPU and openvino_is_available():
                compilers.append(ModelCompiler.OPENVINO)
        return compilers



from tempfile import TemporaryDirectory

import pytest
import torch

from nebullvm.core.models import (
    DeviceType,
    Device,
    DeepLearningFramework,
    QuantizationType,
    ModelCompiler,
)
from nebullvm.operations.inference_learners.torchscript import (
    TorchScriptInferenceLearner,
)
from nebullvm.operations.optimizations.compilers.torchscript import (
    TorchScriptCompiler,
)
from nebullvm.operations.optimizations.optimizers.base import (
    COMPILER_TO_INFERENCE_LEARNER_MAP,
)
from nebullvm.operations.optimizations.tests.utils import (
    initialize_model,
    check_model_validity,
)
from nebullvm.operations.inference_learners.utils import load_model
from nebullvm.tools.utils import gpu_is_available

device = (
    Device(DeviceType.GPU) if gpu_is_available() else Device(DeviceType.CPU)
)


def run_test_torchscript(
    output_library: DeepLearningFramework,
    dynamic: bool,
    quantization_type: QuantizationType,
    metric_drop_ths: int,
    metric: str,
):
    with TemporaryDirectory() as tmp_dir:
        (
            model,
            input_data,
            model_params,
            input_tfms,
            model_outputs,
            metric,
        ) = initialize_model(dynamic, metric, output_library, device)

        compiler_op = TorchScriptCompiler()
        compiler_op.to(device).execute(
            model=model,
            input_tfms=input_tfms,
            metric_drop_ths=metric_drop_ths,
            quantization_type=quantization_type,
            input_data=input_data,
        )

        compiled_model = compiler_op.get_result()

        build_inference_learner_op = COMPILER_TO_INFERENCE_LEARNER_MAP[
            ModelCompiler.TORCHSCRIPT
        ]()

        build_inference_learner_op.to(device).execute(
            model=compiled_model,
            model_orig=compiler_op.model_orig
            if hasattr(compiler_op, "model_orig")
            else None,
            model_params=model_params,
            input_tfms=input_tfms,
            source_dl_framework=output_library,
        )

        optimized_model = build_inference_learner_op.get_result()
        assert isinstance(optimized_model, TorchScriptInferenceLearner)

        # Test save and load functions
        optimized_model.save(tmp_dir)
        loaded_model = load_model(tmp_dir)
        assert isinstance(loaded_model, TorchScriptInferenceLearner)

        assert isinstance(optimized_model.get_size(), int)

        inputs_example = list(optimized_model.get_inputs_example())
        res = optimized_model(*inputs_example)
        assert res is not None

        res_loaded = loaded_model(*inputs_example)
        assert all(
            [
                torch.allclose(res_tensor, res_loaded_tensor)
                for (res_tensor, res_loaded_tensor) in zip(res, res_loaded)
            ]
        )

        # Test validity of the model
        valid = check_model_validity(
            optimized_model,
            input_data,
            model_outputs,
            metric_drop_ths,
            quantization_type,
            metric,
        )
        assert valid

        if dynamic:  # Check also with a smaller bath_size
            torch_device = torch.device(
                "cuda" if torch.cuda.is_available() else "cpu"
            )

            inputs_example = [
                input_[: len(input_) // 2].to(torch_device)
                for input_ in inputs_example
            ]
            res = optimized_model(*inputs_example)
            assert res is not None

            res_orig = tuple(model(*inputs_example))
            assert all(
                [
                    torch.allclose(
                        res_tensor.float(), res_orig_tensor, rtol=2e-01
                    )
                    for (res_tensor, res_orig_tensor) in zip(res, res_orig)
                ]
            )


@pytest.mark.parametrize(
    (
        "output_library",
        "dynamic",
        "quantization_type",
        "metric_drop_ths",
        "metric",
    ),
    [
        (DeepLearningFramework.PYTORCH, True, None, None, None),
        (DeepLearningFramework.PYTORCH, False, None, None, None),
    ],
)
def test_torchscript_no_quantization(
    output_library: DeepLearningFramework,
    dynamic: bool,
    quantization_type: QuantizationType,
    metric_drop_ths: int,
    metric: str,
):
    run_test_torchscript(
        output_library,
        dynamic,
        quantization_type,
        metric_drop_ths,
        metric,
    )


@pytest.mark.parametrize(
    (
        "output_library",
        "dynamic",
        "quantization_type",
        "metric_drop_ths",
        "metric",
    ),
    [
        (
            DeepLearningFramework.PYTORCH,
            True,
            QuantizationType.HALF,
            2,
            "numeric_precision",
        )
    ],
)
@pytest.mark.skipif(
    not torch.cuda.is_available(),
    reason="Half quantization is not available on CPU",
)
def test_torchscript_half_quantization(
    output_library: DeepLearningFramework,
    dynamic: bool,
    quantization_type: QuantizationType,
    metric_drop_ths: int,
    metric: str,
):
    run_test_torchscript(
        output_library,
        dynamic,
        quantization_type,
        metric_drop_ths,
        metric,
    )


@pytest.mark.parametrize(
    (
        "output_library",
        "dynamic",
        "quantization_type",
        "metric_drop_ths",
        "metric",
    ),
    [
        (
            DeepLearningFramework.PYTORCH,
            True,
            QuantizationType.DYNAMIC,
            2,
            "numeric_precision",
        ),
        (
            DeepLearningFramework.PYTORCH,
            True,
            QuantizationType.STATIC,
            2,
            "numeric_precision",
        ),
    ],
)
@pytest.mark.skipif(
    torch.cuda.is_available(),
    reason="INT8 quantization is not available on GPU",
)
def test_torchscript_int8_quantization(
    output_library: DeepLearningFramework,
    dynamic: bool,
    quantization_type: QuantizationType,
    metric_drop_ths: int,
    metric: str,
):
    run_test_torchscript(
        output_library,
        dynamic,
        quantization_type,
        metric_drop_ths,
        metric,
    )



import os

from nebullvm.operations.optimizations.compilers.faster_transformer.bert.modeling_bert import (  # noqa: E501
    BertModel as FasterBertModel,
)
from nebullvm.operations.optimizations.compilers.faster_transformer.bert.modeling_bert import (  # noqa: E501
    CustomEncoder,
    EncoderWeights,
)
from nebullvm.operations.optimizations.compilers.utils import (
    get_faster_transformer_repo_path,
)
from nebullvm.optional_modules.huggingface import BertModel as HFBertModel
from nebullvm.optional_modules.torch import torch


default_lib_path = str(
    get_faster_transformer_repo_path()
    / "build"
    / "lib"
    / "libth_transformer.so"
)


def swap_bert_encoder(model, data_type, lib_path, remove_padding=False):
    """
    Replace the encoder of the model with a custom encoder
    that uses the Faster Transformer library.
    """
    weights = EncoderWeights(
        model.config.num_hidden_layers,
        model.config.hidden_size,
        model.state_dict(),
    )
    weights.to_cuda()
    if data_type == "fp16":
        weights.to_half()
    elif data_type == "bf16":
        weights.to_bfloat16()
    lib_path = os.path.abspath(lib_path)
    enc = CustomEncoder(
        model.config.num_hidden_layers,
        model.config.num_attention_heads,
        model.config.hidden_size // model.config.num_attention_heads,
        weights,
        remove_padding=remove_padding,
        path=lib_path,
    )
    enc_ = torch.jit.script(enc)
    model.replace_encoder(enc_)


def swap_model(
    model: HFBertModel, data_type, lib_path, remove_padding=False
) -> FasterBertModel:
    # bert model need some custom code to call the custom encoder
    # so we need to use custom bert class
    new_model = FasterBertModel(model.config)
    new_model.load_state_dict(model.state_dict())
    swap_bert_encoder(new_model, data_type, lib_path, remove_padding)
    return new_model


def detect_and_swap_bert_model(
    model, data_type, lib_path=default_lib_path, remove_padding=False
):
    if type(model) == HFBertModel:
        model = swap_model(model, data_type, lib_path, remove_padding)
    if hasattr(model, "bert") and type(model.bert) == HFBertModel:
        model.bert = swap_model(
            model.bert, data_type, lib_path, remove_padding
        )
    return model



from nebullvm.optional_modules.dummy import DummyClass

try:
    import torch_tensorrt
    from torch_tensorrt.ptq import DataLoaderCalibrator  # noqa F401
except ImportError:
    torch_tensorrt = DummyClass
    DataLoaderCalibrator = None



import multiprocessing
import os
import shutil
from abc import ABC
from pathlib import Path
from typing import Union, List, Generator, Tuple, Dict, Type

import cpuinfo
import numpy as np
from loguru import logger

from nebullvm.config import (
    ONNX_FILENAMES,
    ONNX_PROVIDERS,
)
from nebullvm.core.models import (
    QuantizationType,
    Device,
    DeviceType,
    ModelParams,
    DeepLearningFramework,
)
from nebullvm.operations.inference_learners.base import (
    BaseInferenceLearner,
    LearnerMetadata,
    PytorchBaseInferenceLearner,
    TensorflowBaseInferenceLearner,
    NumpyBaseInferenceLearner,
)
from nebullvm.operations.optimizations.compilers.utils import (
    tensorrt_is_available,
)
from nebullvm.optional_modules.onnx import onnx
from nebullvm.optional_modules.onnxruntime import onnxruntime as ort
from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import torch
from nebullvm.tools.transformations import MultiStageTransformation


def _running_on_intel_cpu(use_gpu):
    if use_gpu:
        return False  # running on GPU
    cpu_info = cpuinfo.get_cpu_info()["brand_raw"].lower()
    if "intel" in cpu_info:
        return True
    return False


def _get_ort_session_options(use_gpu) -> ort.SessionOptions:
    sess_options = ort.SessionOptions()
    sess_options.graph_optimization_level = (
        ort.GraphOptimizationLevel.ORT_ENABLE_ALL
    )
    if not use_gpu:
        sess_options.execution_mode = ort.ExecutionMode.ORT_PARALLEL
        sess_options.inter_op_num_threads = 1
        sess_options.intra_op_num_threads = max(
            int(
                os.environ.get("NEBULLVM_THREADS_PER_MODEL")
                or multiprocessing.cpu_count()
            ),
            1,
        )
    return sess_options


class ONNXInferenceLearner(BaseInferenceLearner, ABC):
    """Model converted to ONNX and run with Microsoft's onnxruntime.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        onnx_path (str or Path): Path to the onnx model.
        input_names (List[str]): Input names used when the onnx model
            was produced.
        output_names (List[str]): Output names used when the onnx model
            was produced.
    """

    name = "ONNXRuntime"

    def __init__(
        self,
        onnx_path: Union[str, Path],
        input_names: List[str],
        output_names: List[str],
        device: Device,
        quantization_type: QuantizationType,
        **kwargs,
    ):
        super().__init__(**kwargs)
        filename = Path(onnx_path).name
        dir_path = str(Path(onnx_path).parent)
        self.device = device

        self.onnx_path = Path(self._store_dir(dir_path)) / filename
        self.sess_options = _get_ort_session_options(
            self.device.type is DeviceType.GPU
        )
        self.quantization_type = quantization_type

        if _running_on_intel_cpu(self.device.type is DeviceType.GPU):
            self.sess_options.add_session_config_entry(
                "session.set_denormal_as_zero", "1"
            )

        self.set_model_on_gpu()

        self._is_gpu_ready = self.device.type is DeviceType.GPU
        self.input_names = input_names
        self.output_names = output_names

    @staticmethod
    def _setup_tensorrt(quantization_type: QuantizationType, device: Device):
        if (
            tensorrt_is_available()
            and os.environ.get("LD_LIBRARY_PATH", False)
            and "tensorrt" in os.environ["LD_LIBRARY_PATH"]
        ):
            ONNX_PROVIDERS["cuda"][0] = (
                "TensorrtExecutionProvider",
                {
                    "device_id": device.idx,
                    "trt_max_workspace_size": device.get_free_memory(),
                    "trt_fp16_enable": True
                    if quantization_type is not None
                    else False,
                    "trt_int8_enable": True
                    if quantization_type is QuantizationType.STATIC
                    else False,
                },
            )
        else:
            if tensorrt_is_available():
                logger.warning(
                    "TensorrtExecutionProvider for onnx is not "
                    "available. If you want to use it, please  "
                    "add the path to tensorrt to the "
                    "LD_LIBRARY_PATH environment variable. "
                    "CUDA provider will be used instead. "
                )
            else:
                logger.warning(
                    "TensorRT is not available. "
                    "If you want to use it, please install it and "
                    "add the path to the LD_LIBRARY_PATH "
                    "environment variable."
                    "CUDA provider will be used instead. "
                )
            if "TensorrtExecutionProvider" in ONNX_PROVIDERS["cuda"]:
                ONNX_PROVIDERS["cuda"].remove("TensorrtExecutionProvider")

    def get_size(self):
        return sum(
            os.path.getsize(self.onnx_path.parents[0] / f)
            for f in os.listdir(self.onnx_path.parents[0])
            if os.path.isfile(self.onnx_path.parents[0] / f)
        )

    def free_gpu_memory(self):
        del self._session
        self._is_gpu_ready = False

    def set_model_on_gpu(self):
        if (
            self.device.type is DeviceType.GPU
            and len(ONNX_PROVIDERS["cuda"]) == 3
        ):
            ONNX_PROVIDERS["cuda"][1] = (
                "CUDAExecutionProvider",
                {
                    "device_id": self.device.idx,
                },
            )
            self._setup_tensorrt(self.quantization_type, self.device)

        ort_session = ort.InferenceSession(
            str(self.onnx_path),
            sess_options=self.sess_options,
            providers=ONNX_PROVIDERS["cuda"]
            if self.device.type is DeviceType.GPU
            else ONNX_PROVIDERS["cpu"],
        )
        self._session = ort_session
        self._is_gpu_ready = True

    def save(self, path: Union[str, Path], **kwargs):
        """Save the model.

        Args:
            path (Path or str): Path to the directory where the model will
                be stored.
            kwargs (Dict): Dictionary of key-value pairs that will be saved in
                the model metadata file.
        """
        metadata = LearnerMetadata.from_model(
            self,
            input_names=self.input_names,
            output_names=self.output_names,
            **kwargs,
        )

        path = Path(path)
        path.mkdir(exist_ok=True)

        metadata.save(path)

        shutil.copy(
            self.onnx_path,
            os.path.join(str(path), ONNX_FILENAMES["model_name"]),
        )

        try:
            # Tries to load the model
            onnx.load(os.path.join(str(path), ONNX_FILENAMES["model_name"]))
        except FileNotFoundError:
            # If missing files, it means it's saved in onnx external_data
            # format
            src_dir = str(Path(self.onnx_path).parent)
            files = os.listdir(src_dir)
            for fname in files:
                if ".onnx" not in fname:
                    shutil.copy2(
                        os.path.join(src_dir, fname), os.path.join(path, fname)
                    )

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs):
        """Load the model.

        Args:
            path (Path or str): Path to the directory where the model is
                stored.
            kwargs (Dict): Dictionary of additional arguments for consistency
                with other Learners.

        Returns:
            ONNXInferenceLearner: The optimized model.
        """
        if len(kwargs) > 0:
            logger.warning(
                f"No extra keywords expected for the load method. "
                f"Got {kwargs}."
            )
        path = Path(path)
        onnx_path = path / ONNX_FILENAMES["model_name"]
        metadata = LearnerMetadata.read(path)
        input_tfms = metadata.input_tfms
        device = Device.from_str(metadata.device)
        quantization_type = (
            QuantizationType(metadata.quantization_type)
            if hasattr(metadata, "quantization_type")
            else None
        )
        if input_tfms is not None:
            input_tfms = MultiStageTransformation.from_dict(
                metadata.input_tfms
            )
        return cls(
            input_tfms=input_tfms,
            network_parameters=ModelParams(**metadata.network_parameters),
            onnx_path=onnx_path,
            input_names=metadata["input_names"],
            output_names=metadata["output_names"],
            device=device,
            quantization_type=quantization_type,
        )

    def _predict_arrays(self, input_arrays: Generator[np.ndarray, None, None]):
        input_dict = {
            name: input_array
            for name, input_array in zip(self.input_names, input_arrays)
        }
        outputs = self._session.run(self.output_names, input_dict)
        return outputs


class PytorchONNXInferenceLearner(
    ONNXInferenceLearner, PytorchBaseInferenceLearner
):
    """Model run with Microsoft's onnxruntime using a Pytorch interface.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        onnx_path (str or Path): Path to the onnx model.
        input_names (List[str]): Input names used when the onnx model
            was produced.
        output_names (List[str]): Output names used when the onnx model
            was produced.
    """

    def run(self, *input_tensors: torch.Tensor) -> Tuple[torch.Tensor]:
        """Predict on the input tensors.

        Note that the input tensors must be on the same batch. If a sequence
        of tensors is given when the model is expecting a single input tensor
        (with batch size >= 1) an error is raised.

        Args:
            input_tensors (Tuple[Tensor]): Input tensors belonging to the same
                batch. The tensors are expected having dimensions
                (batch_size, dim1, dim2, ...).

        Returns:
            Tuple[Tensor]: Output tensors. Note that the output tensors does
                not correspond to the prediction on the input tensors with a
                1 to 1 mapping. In fact the output tensors are produced as the
                multiple-output of the model given a (multi-) tensor input.
        """
        if self.device.type is DeviceType.GPU and not self._is_gpu_ready:
            self.set_model_on_gpu()
        input_arrays = (
            input_tensor.cpu().detach().numpy()
            for input_tensor in input_tensors
        )
        outputs = self._predict_arrays(input_arrays)
        return tuple(
            torch.from_numpy(output).to(self.device.to_torch_format())
            for output in outputs
        )


class TensorflowONNXInferenceLearner(
    ONNXInferenceLearner, TensorflowBaseInferenceLearner
):
    """Model run with Microsoft's onnxruntime using a tensorflow interface.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        onnx_path (str or Path): Path to the onnx model.
        input_names (List[str]): Input names used when the onnx model
            was produced.
        output_names (List[str]): Output names used when the onnx model
            was produced.
    """

    def run(self, *input_tensors: tf.Tensor) -> Tuple[tf.Tensor, ...]:
        """Predict on the input tensors.

        Note that the input tensors must be on the same batch. If a sequence
        of tensors is given when the model is expecting a single input tensor
        (with batch size >= 1) an error is raised.

        Args:
            input_tensors (Tuple[Tensor]): Input tensors belonging to the same
                batch. The tensors are expected having dimensions
                (batch_size, dim1, dim2, ...).

        Returns:
            Tuple[Tensor]: Output tensors. Note that the output tensors does
                not correspond to the prediction on the input tensors with a
                1 to 1 mapping. In fact the output tensors are produced as the
                multiple-output of the model given a (multi-) tensor input.
        """
        if self.device.type is DeviceType.GPU and not self._is_gpu_ready:
            self.set_model_on_gpu()
        input_arrays = (
            input_tensor.numpy()
            if not isinstance(input_tensor, np.ndarray)
            else input_tensor
            for input_tensor in input_tensors
        )
        outputs = self._predict_arrays(input_arrays)
        # noinspection PyTypeChecker
        return tuple(tf.convert_to_tensor(output) for output in outputs)


class NumpyONNXInferenceLearner(
    ONNXInferenceLearner, NumpyBaseInferenceLearner
):
    """Model run with Microsoft's onnxruntime using a numpy interface.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        onnx_path (str or Path): Path to the onnx model.
        input_names (List[str]): Input names used when the onnx model
            was produced.
        output_names (List[str]): Output names used when the onnx model
            was produced.
    """

    def run(self, *input_tensors: np.ndarray) -> Tuple[np.ndarray, ...]:
        """Predict on the input tensors.

        Note that the input tensors must be on the same batch. If a sequence
        of tensors is given when the model is expecting a single input tensor
        (with batch size >= 1) an error is raised.

        Args:
            input_tensors (Tuple[np.ndarray, ...]): Input tensors belonging to
                the same batch. The tensors are expected having dimensions
                (batch_size, dim1, dim2, ...).

        Returns:
            Tuple[Tensor]: Output tensors. Note that the output tensors does
                not correspond to the prediction on the input tensors with a
                1 to 1 mapping. In fact the output tensors are produced as the
                multiple-output of the model given a (multi-) tensor input.
        """
        if self.device.type is DeviceType.GPU and not self._is_gpu_ready:
            self.set_model_on_gpu()
        input_arrays = (input_tensor for input_tensor in input_tensors)
        outputs = self._predict_arrays(input_arrays)
        return tuple(outputs)


ONNX_INFERENCE_LEARNERS: Dict[
    DeepLearningFramework, Type[ONNXInferenceLearner]
] = {
    DeepLearningFramework.PYTORCH: PytorchONNXInferenceLearner,
    DeepLearningFramework.TENSORFLOW: TensorflowONNXInferenceLearner,
    DeepLearningFramework.NUMPY: NumpyONNXInferenceLearner,
}



# Based on: https://github.com/NVIDIA/FasterTransformer/blob/4402759e48f2340220638675f464b6ba1f79ac3c/examples/pytorch/gpt/gpt_summarization.py # noqa: E501
# Copyright (c) 2022-2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile
from typing import Callable, Iterable, List, Optional, Tuple, Union

from nebullvm.operations.optimizations.compilers.faster_transformer.gpt.utils import \
    gpt_decoder
from nebullvm.operations.optimizations.compilers.faster_transformer.gpt.utils.huggingface_gpt_convert import (  # noqa: E501
    main as convert_huggingface_gpt_to_faster_transformer,
)
from nebullvm.operations.optimizations.compilers.utils import (
    get_faster_transformer_repo_path,
)
from nebullvm.optional_modules.huggingface import GPT2LMHeadModel
from nebullvm.optional_modules.torch import torch

lib_path = default_lib_path = str(
    get_faster_transformer_repo_path()
    / "build"
    / "lib"
    / "libth_transformer.so"
)


class FasterTransformerGPT2Wrapper(torch.nn.Module):
    def __init__(self, model: gpt_decoder.Gpt, config):
        super().__init__()
        self.model = model
        self.config = config
        self.device = model.device

    @torch.no_grad()
    def generate(
        self,
        inputs: Optional[torch.Tensor] = None,
        max_length: Optional[int] = None,
        min_length: Optional[int] = None,
        do_sample: Optional[bool] = None,
        early_stopping: Optional[bool] = None,
        num_beams: Optional[int] = 1,
        temperature: Optional[float] = None,
        penalty_alpha: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        typical_p: Optional[float] = None,
        repetition_penalty: Optional[float] = None,
        bad_words_ids: Optional[Iterable[int]] = None,
        force_words_ids: Optional[
            Union[Iterable[int], Iterable[Iterable[int]]]
        ] = None,
        bos_token_id: Optional[int] = None,
        pad_token_id: Optional[int] = None,
        eos_token_id: Optional[int] = None,
        length_penalty: Optional[float] = None,
        no_repeat_ngram_size: Optional[int] = None,
        encoder_no_repeat_ngram_size: Optional[int] = None,
        num_return_sequences: Optional[int] = None,
        max_time: Optional[float] = None,
        max_new_tokens: Optional[int] = None,
        decoder_start_token_id: Optional[int] = None,
        use_cache: Optional[bool] = None,
        num_beam_groups: Optional[int] = None,
        diversity_penalty: Optional[float] = None,
        prefix_allowed_tokens_fn: Optional[
            Callable[[int, torch.Tensor], List[int]]
        ] = None,
        # logits_processor: Optional[LogitsProcessorList] = None,
        # renormalize_logits: Optional[bool] = None,
        # stopping_criteria: Optional[StoppingCriteriaList] = None,
        # constraints: Optional[List[Constraint]] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        output_scores: Optional[bool] = None,
        return_dict_in_generate: Optional[bool] = None,
        forced_bos_token_id: Optional[int] = None,
        forced_eos_token_id: Optional[int] = None,
        remove_invalid_values: Optional[bool] = None,
        synced_gpus: Optional[bool] = False,
        exponential_decay_length_penalty: Optional[Tuple[int, float]] = None,
        suppress_tokens: Optional[List[int]] = None,
        begin_suppress_tokens: Optional[List[int]] = None,
        forced_decoder_ids: Optional[List[List[int]]] = None,
    ):

        input_lengths = torch.tensor(
            [len(input) for input in inputs],
            dtype=torch.int32,
            device=self.model.device,
        )
        batch_size = len(inputs)

        def convert_to_tensor_if_not(value, dtype=torch.float32):
            if value is None:
                return value
            if isinstance(value, torch.Tensor):
                return value
            return value * torch.ones(batch_size, dtype=dtype)  # cpu tensor

        top_k = convert_to_tensor_if_not(top_k, dtype=torch.int32)
        top_p = convert_to_tensor_if_not(top_p, dtype=torch.float32)
        temperature = convert_to_tensor_if_not(
            temperature, dtype=torch.float32
        )
        repetition_penalty = convert_to_tensor_if_not(
            repetition_penalty, dtype=torch.float32
        )
        min_length = convert_to_tensor_if_not(min_length, dtype=torch.int32)
        len_penalty = convert_to_tensor_if_not(
            length_penalty, dtype=torch.float32
        )
        if max_length is None:
            # gen_length is required for faster transformer
            # infer it from the model config
            max_length = self.config.n_ctx
        output_dict = self.model.generate(
            input_token_ids=inputs,
            input_lengths=input_lengths,
            gen_length=max_length - len(inputs[0]),
            eos_token_id=eos_token_id,
            # local_batch_size=None,
            beam_width=num_beams,
            top_k=top_k,
            top_p=top_p,
            # top_p_decay: Optional[torch.FloatTensor] = None,
            # top_p_min: Optional[torch.FloatTensor] = None,
            # top_p_reset_ids: Optional[torch.IntTensor] = None,
            temperature=temperature,
            repetition_penalty=repetition_penalty,
            # presence_penalty: Optional[torch.FloatTensor] = None,
            min_length=min_length,
            len_penalty=len_penalty,
            # beam_search_diversity_rate: Optional[torch.FloatTensor] = None,
            # stop_words_list: Optional[torch.IntTensor] = None,
            # bad_words_list: Optional[torch.IntTensor] = None,
            # sequence_limit_lengths: Optional[torch.IntTensor] = None,
            # random_seed: Optional[torch.LongTensor] = None,
            # memory_length: Optional[int] = None,
            return_output_length=True,
            return_log_probs=False,
        )
        output_token_ids = output_dict["output_token_ids"]
        output_lengths = output_dict["output_lengths"]
        # tokens = output_token_ids[0, 0, input_lengths[0]:output_lengths[0]]
        tokens = [
            # output_token_ids[i, 0, input_lengths[i]:output_lengths[i]]
            output_token_ids[i, 0, : output_lengths[i]]
            for i in range(batch_size)
        ]
        return tokens


def convert_gpt2_lm_head_model(
    model: GPT2LMHeadModel,
    tokenizer,
    weight_data_type="fp32",
    data_type="fp16",
    use_fp32_to_compute_logit=False,
):
    """
    currently doens't support fp8 or multi-gpu
    """
    weights_data_type = weight_data_type
    temp_dir = tempfile.TemporaryDirectory()
    temp_dir_path = temp_dir.name
    ft_model_location = saved_dir = temp_dir_path + "/gpt2"
    hf_config = model.config.to_dict()
    # convert huggingface model to faster transformer model
    convert_huggingface_gpt_to_faster_transformer(
        saved_dir=saved_dir,
        model=model.transformer,
        weight_data_type=weight_data_type,
    )

    head_num = hf_config["n_head"]
    layer_num = hf_config["n_layer"]
    start_id = hf_config["bos_token_id"]
    end_id = hf_config["eos_token_id"]
    size_per_head = hf_config["n_embd"] // head_num

    vocab_size = tokenizer.vocab_size

    tensor_para_size = 1
    pipeline_para_size = 1
    ckpt_path = os.path.join(ft_model_location, f"{tensor_para_size}-gpu")
    max_seq_len = hf_config["n_ctx"]
    int8_mode = 0  # 0: no quantization, 1: quantize weights to int8
    # load faster transformer model, note that the lm_head is not saved
    # it's reconstructed during loading from the embedding weights
    gpt = gpt_decoder.Gpt(
        num_heads=head_num,
        size_per_head=size_per_head,
        num_layers=layer_num,
        vocab_size=vocab_size,
        start_id=start_id,
        end_id=end_id,
        tensor_para_size=tensor_para_size,
        pipeline_para_size=pipeline_para_size,
        lib_path=lib_path,
        max_seq_len=max_seq_len,
        int8_mode=int8_mode,
        inference_data_type=data_type,
        weights_data_type=weights_data_type,
        use_fp32_to_compute_logit=use_fp32_to_compute_logit,
    )
    gpt.load(ckpt_path, data_type)
    return FasterTransformerGPT2Wrapper(gpt, model.config)


# from transformers import GPT2LMHeadModel, GPT2Tokenizer
# tokenizer = GPT2Tokenizer.from_pretrained("gpt2")
# tokenizer.pad_token = tokenizer.eos_token
# model = hf_model = GPT2LMHeadModel.from_pretrained("gpt2").to("cuda").eval()
# hf_config = hf_model.config.to_dict()


# model = GPT2LMHeadModel.from_pretrained("gpt2")
# tokenizer = GPT2Tokenizer.from_pretrained('gpt2')
# weight_data_type = weights_data_type = "fp32" # fp32 or fp16
# data_type = "fp32" # fp32 or fp16
# faster_model= convert_gpt2_lm_head_model(
# model, tokenizer,
# weight_data_type=weight_data_type,
# data_type=data_type)



from pathlib import Path
from typing import List, Tuple, Optional, Dict, Union, Sequence

from loguru import logger

from nebullvm.core.models import Device, DataType, DeviceType, InputInfo
from nebullvm.optional_modules.torch import torch, DataLoader
from nebullvm.tools.data import DataManager
from nebullvm.tools.diffusers import get_default_dynamic_info

FX_MODULE_NAME = "NebullvmFxModule"


def save_with_torch_fx(model: torch.nn.Module, path: Path):
    traced_model = torch.fx.symbolic_trace(model)
    traced_model.to_folder(path, FX_MODULE_NAME)


def load_with_torch_fx(
    path: Path, state_dict_name: str = "pruned_state_dict.pt"
):
    module_file = path / "module.py"
    with open(module_file, "r") as f:
        module_str = f.read()
    exec(module_str, globals())
    model = eval(FX_MODULE_NAME)()
    model.load_state_dict(torch.load(path / state_dict_name))
    return model


def get_output_info_torch(
    torch_model: torch.nn.Module,
    input_tensors: List[torch.Tensor],
    device: Device,
) -> List[Tuple[Tuple[int, ...], DataType]]:
    if device.type is DeviceType.GPU:
        input_tensors = [x.to(device.to_torch_format()) for x in input_tensors]
        torch_model.to(device.to_torch_format())
    with torch.no_grad():
        outputs = torch_model(*input_tensors)
        if isinstance(outputs, torch.Tensor):
            return [
                (
                    tuple(outputs.size()),
                    DataType.from_framework_format(outputs.dtype),
                )
            ]
        else:
            return [
                (
                    tuple(output.size()),
                    DataType.from_framework_format(output.dtype),
                )
                for output in outputs
            ]


def create_model_inputs_torch(
    input_infos: List[InputInfo],
) -> List[torch.Tensor]:
    input_tensors = (
        torch.randn(*input_info.size)
        if input_info.dtype is DataType.FLOAT32
        else torch.randint(
            size=input_info.size,
            low=input_info.min_value or 0,
            high=input_info.max_value or 100,
        )
        for input_info in input_infos
    )
    return list(input_tensors)


def run_torch_model(
    torch_model: torch.nn.Module,
    input_tensors: List[torch.Tensor],
    device: Device,
    dtype: torch.dtype = torch.float,
) -> List[torch.Tensor]:
    torch_model.eval()
    if device.type is DeviceType.GPU:
        torch_model.to(device.to_torch_format())
        if dtype != torch.half:
            input_tensors = (
                t.to(device.to_torch_format()) for t in input_tensors
            )
        else:
            input_tensors = (
                t.to(device.to_torch_format()).half()
                if t.dtype == torch.float
                else t.to(device.to_torch_format())
                for t in input_tensors
            )
    with torch.no_grad():
        pred = torch_model(*input_tensors)
    if isinstance(pred, torch.Tensor):
        pred = [pred.cpu()]
    else:
        pred = [p.cpu() for p in pred]
    return pred


def _extract_dynamic_axis(
    torch_model: torch.nn.Module,
    dataloader: DataManager,
    input_sizes: List[Tuple[int, ...]],
    device: Device,
    max_data: int = 100,
) -> Optional[Dict]:
    from nebullvm.tools.utils import inspect_dynamic_size

    dynamic_axis = {"inputs": [{}] * len(input_sizes), "outputs": []}
    output_sizes = []
    for i, input_data in enumerate(dataloader):
        input_tensors = input_data[0]
        if i >= max_data:
            break
        inspect_dynamic_size(
            input_tensors, input_sizes, dynamic_axis["inputs"]
        )
        outputs = tuple(run_torch_model(torch_model, input_tensors, device))
        if i == 0:
            dynamic_axis["outputs"] = [{}] * len(outputs)
            output_sizes = [tuple(output.shape) for output in outputs]
        inspect_dynamic_size(outputs, output_sizes, dynamic_axis["outputs"])
    if any(
        len(x) > 0 for x in (dynamic_axis["inputs"] + dynamic_axis["outputs"])
    ):
        return dynamic_axis
    return None


def extract_info_from_torch_data(
    model: torch.nn.Module,
    dataloader: Union[DataLoader, Sequence],
    dynamic_axis: Dict,
    device: Device,
    is_diffusion: bool = False,
):
    from nebullvm.tools.utils import ifnone

    input_data = (
        dataloader[0]
        if isinstance(dataloader, Sequence)
        else next(iter(dataloader))
    )
    input_row = input_data[0]
    batch_size = int(input_row[0].shape[0])
    if not all([input_row[0].shape[0] == x.shape[0] for x in input_row]):
        logger.warning("Detected not consistent batch size in the inputs.")

    input_sizes = [tuple(x.shape) for x in input_row]
    input_types = [
        "int64"
        if isinstance(x.cpu(), torch.LongTensor)
        else "int32"
        if isinstance(x.cpu(), torch.IntTensor)
        else "float16"
        if isinstance(x.cpu(), torch.HalfTensor)
        else "float32"
        for x in input_row
    ]

    # For the Stable Diffusion UNet we must provide dynamic axis
    # even when using static shapes, because otherwise the converted
    # onnx model will have size issues.
    if dynamic_axis is None and device.type is DeviceType.GPU and is_diffusion:
        dynamic_axis = get_default_dynamic_info(input_sizes)

    if dynamic_axis is not None:
        dynamic_axis["inputs"] = [
            {int(k): v for (k, v) in val.items()}
            for val in dynamic_axis["inputs"]
        ]
        dynamic_axis["outputs"] = [
            {int(k): v for (k, v) in val.items()}
            for val in dynamic_axis["outputs"]
        ]

    dynamic_axis = ifnone(
        dynamic_axis,
        _extract_dynamic_axis(model, dataloader, input_sizes, device),
    )
    return batch_size, input_sizes, input_types, dynamic_axis


def torch_is_gpu_available():
    return torch.cuda.is_available()


def torch_get_device_name():
    return torch.cuda.get_device_name(0)


def get_torch_model_size(
    model: Union[torch.nn.Module, torch.jit.ScriptModule, torch.fx.GraphModule]
):
    param_size = 0
    for param in model.parameters():
        param_size += param.nelement() * param.element_size()
    buffer_size = 0
    for buffer in model.buffers():
        buffer_size += buffer.nelement() * buffer.element_size()

    return param_size + buffer_size



import os
import subprocess
import sys
import uuid
from pathlib import Path
from types import ModuleType
from typing import (
    Tuple,
    Any,
    List,
    Dict,
    Union,
    Iterable,
    Sequence,
    Optional,
    Callable,
)

import numpy as np
from loguru import logger
from packaging import version

from nebullvm.core.models import (
    DeepLearningFramework,
    Device,
    ModelParams,
    DeviceType,
)
from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import torch
from nebullvm.tools.data import DataManager
from nebullvm.tools.onnx import (
    extract_info_from_np_data,
    get_output_info_onnx,
)
from nebullvm.tools.pytorch import (
    extract_info_from_torch_data,
    get_output_info_torch,
)
from nebullvm.tools.tf import (
    extract_info_from_tf_data,
    get_output_info_tf,
)


def get_model_size_mb(model: Any) -> float:
    if isinstance(model, str):
        size = os.stat(model).st_size
    elif isinstance(model, Path):
        size = os.path.getsize(model.as_posix())
    elif isinstance(model, torch.nn.Module):
        size = sum(p.nelement() * p.element_size() for p in model.parameters())
    else:
        # we assume it is a tf_model
        # assuming full precision 32 bit
        size = model.count_params() * 4
    return round(size * 1e-6, 2)


def get_model_name(model: Any) -> str:
    if isinstance(model, str):
        return model
    if isinstance(model, Path):
        return model.as_posix()
    return model.__class__.__name__


def generate_model_id(model: Any) -> str:
    model_name = get_model_name(model)
    return f"{str(uuid.uuid4())}_{hash(model_name)}"


def get_throughput(latency: float, batch_size: int = 1) -> float:
    if latency == 0:
        return -1
    return (1 / latency) * batch_size


def ifnone(target, new_value):
    if target is None:
        return new_value
    else:
        return target


def inspect_dynamic_size(
    tensors: Tuple[Any, ...],
    sizes: List[Tuple[int, ...]],
    axis_list: List[Dict],
):
    for idx, (tensor, size) in enumerate(zip(tensors, sizes)):
        for idy, (j, k) in enumerate(zip(tensor.shape, size)):
            if j != k:
                if idy == 0:
                    tag = "batch_size"
                else:
                    tag = f"val_{j}_{k}"
                axis_list[idx][idy] = tag


def gpu_is_available():
    try:
        subprocess.check_output("nvidia-smi")
        return True
    except Exception:
        return False


def neuron_is_available():
    try:
        subprocess.check_output("neuron-ls")
        return True
    except Exception:
        return False


def tpu_is_available():
    # Check if a tpu is available
    try:
        import torch_xla
        import torch_xla.core.xla_model as xm

        return xm.xla_device_hw(torch_xla.core.xla_model.xla_device()) == "TPU"
    except Exception:
        return False


def check_module_version(
    module: ModuleType, min_version: str = None, max_version: str = None
) -> bool:
    installed_version = module.__version__

    if min_version is not None:
        if version.parse(installed_version) < version.parse(min_version):
            return False

    if max_version is not None:
        if version.parse(installed_version) > version.parse(max_version):
            return False

    return True


def is_python_version_3_10():
    return (
        str(sys.version_info.major) + "." + str(sys.version_info.minor)
        == "3.10"
    )


def get_dl_framework(model: Any):
    if isinstance(model, torch.nn.Module):
        return DeepLearningFramework.PYTORCH
    elif isinstance(model, tf.Module) and model is not None:
        return DeepLearningFramework.TENSORFLOW
    elif isinstance(model, str):
        if Path(model).is_file():
            return DeepLearningFramework.NUMPY
        else:
            raise FileNotFoundError(
                f"No file '{model}' found, please provide a valid path to "
                f"a model."
            )
    else:
        raise TypeError(f"Model type {type(model)} not supported.")


def check_input_data(input_data: Union[Iterable, Sequence]):
    try:
        assert len(input_data) > 0
        assert isinstance(input_data[0], tuple)
        assert isinstance(input_data[0][0], tuple)
        assert isinstance(
            input_data[0][0][0], (np.ndarray, torch.Tensor, tf.Tensor)
        )
        if len(input_data[0]) > 1:
            assert isinstance(
                input_data[0][1],
                (np.ndarray, torch.Tensor, tf.Tensor, int, float, type(None)),
            )
    except:  # noqa E722
        return False
    else:
        return True


def is_data_subscriptable(input_data: Union[Iterable, Sequence]):
    try:
        input_data[0]
    except:  # noqa E722
        return False
    else:
        return True


def check_dynamic_info_inputs(
    dynamic_info: Optional[Dict], input_sample: Tuple[Any]
):
    if dynamic_info is not None:
        assert dynamic_info.get("inputs") is not None, (
            "Dynamic info must contain an 'inputs' key with a list of "
            "dictionaries as value."
        )

        num_dynamic_inputs = len(dynamic_info["inputs"])
        num_model_inputs = len(input_sample)
        assert len(dynamic_info["inputs"]) == len(input_sample), (
            f"The number of dynamic inputs provided in the dynamic info "
            f"dict ({num_dynamic_inputs}) is not equal to the number "
            f"of inputs of the model ({num_model_inputs}). Detected model "
            f"input shapes are: {[input.shape for input in input_sample]} "
        )

        assert dynamic_info.get("outputs") is not None, (
            "Dynamic info must contain an 'outputs' key with a list of "
            "dictionaries as value."
        )


def extract_info_from_data(
    model: Any,
    input_data: DataManager,
    dl_framework: DeepLearningFramework,
    dynamic_info: Optional[Dict],
    device: Device,
    is_diffusion: bool = False,
):
    check_dynamic_info_inputs(dynamic_info, input_data.get_list(1)[0])
    batch_size, input_sizes, input_types, dynamic_info = INFO_EXTRACTION_DICT[
        dl_framework
    ](
        model,
        input_data,
        dynamic_axis=dynamic_info,
        device=device,
        is_diffusion=is_diffusion,
    )

    output_infos = OUTPUT_INFO_COMPUTATION_DICT[dl_framework](
        model, input_data[0][0], device
    )
    model_params = ModelParams(
        batch_size=batch_size,
        input_infos=[
            {"size": size, "dtype": dtype}
            for size, dtype in zip(input_sizes, input_types)
        ],
        output_sizes=[info[0] for info in output_infos],
        output_types=[info[1] for info in output_infos],
        dynamic_info=dynamic_info,
    )
    return model_params


def is_huggingface_data(data_sample: Any) -> bool:
    if is_dict_type(data_sample):
        return True
    elif isinstance(data_sample, str):
        return True
    elif isinstance(data_sample[0], str):
        return True
    return False


def is_dict_type(data_sample: Any):
    try:
        data_sample.items()
    except AttributeError:
        return False
    else:
        return True


def _get_idx(device: str) -> int:
    device_info = device.split(":")
    if len(device_info) == 2 and device_info[1].isdigit():
        idx = int(device_info[1])
    else:
        idx = 0
    return idx


def _set_device(
    accelerator_is_available: bool, device_type: DeviceType, idx: int
) -> Device:
    if not accelerator_is_available:
        logger.warning(
            f"Selected {device_type.name} device but no available "
            f"{device_type.name} found on this platform. CPU will "
            f"be used instead. Please make sure that the "
            f"{device_type.name} is installed and can be used by your "
            "framework."
        )
        device = Device(DeviceType.CPU)
    else:
        device = Device(device_type, idx=idx)

    return device


def check_device(device: Optional[str] = None) -> Device:
    if device is None:
        if gpu_is_available():
            device = Device(DeviceType.GPU)
        elif neuron_is_available():
            device = Device(DeviceType.NEURON)
        elif tpu_is_available():
            device = Device(DeviceType.TPU)
        else:
            device = Device(DeviceType.CPU)
    else:
        if any(x in device.lower() for x in ["cuda", "gpu"]):
            device = _set_device(
                accelerator_is_available=gpu_is_available(),
                device_type=DeviceType.GPU,
                idx=_get_idx(device),
            )
        elif "neuron" in device.lower():
            device = _set_device(
                accelerator_is_available=neuron_is_available(),
                device_type=DeviceType.NEURON,
                idx=_get_idx(device),
            )
        elif "tpu" in device.lower():
            device = _set_device(
                accelerator_is_available=tpu_is_available(),
                device_type=DeviceType.TPU,
                idx=_get_idx(device),
            )
        else:
            device = Device(DeviceType.CPU)

    return device


def get_gpu_compute_capability(gpu_idx: int) -> float:
    compute_capability = subprocess.check_output(
        ["nvidia-smi", "--query-gpu=compute_cap", "--format=csv,noheader"]
    ).decode("utf-8")
    return float(compute_capability.split("\n")[gpu_idx])


INFO_EXTRACTION_DICT: Dict[DeepLearningFramework, Callable] = {
    DeepLearningFramework.PYTORCH: extract_info_from_torch_data,
    DeepLearningFramework.TENSORFLOW: extract_info_from_tf_data,
    DeepLearningFramework.NUMPY: extract_info_from_np_data,
}

OUTPUT_INFO_COMPUTATION_DICT: Dict[DeepLearningFramework, Callable] = {
    DeepLearningFramework.PYTORCH: get_output_info_torch,
    DeepLearningFramework.TENSORFLOW: get_output_info_tf,
    DeepLearningFramework.NUMPY: get_output_info_onnx,
}



from pathlib import Path
from typing import Union, Any

from nebullvm.operations.inference_learners.base import LearnerMetadata
from nebullvm.optional_modules.diffusers import StableDiffusionPipeline
from nebullvm.tools.diffusers import postprocess_diffusers


def load_model(path: Union[Path, str], pipe: StableDiffusionPipeline = None):
    """Load the optimized model previously saved in the given path.

    Args:
        path (Union[Path, str]): Path to the directory where the model is
            saved.
        pipe (StableDiffusionPipeline): Diffusion pipeline to be used for
            loading the model. This parameter is only needed if the model
            to be loaded is a diffusion model. Default: None.

    Returns:
        InferenceLearner: Model optimized by Speedster.
    """
    optimized_model = LearnerMetadata.read(path).load_model(path)
    if pipe is not None:
        optimized_model = postprocess_diffusers(
            optimized_model, pipe, optimized_model.device
        )
    return optimized_model


def save_model(model: Any, path: Union[Path, str]):
    """Save the optimized model in the given path.

    Args:
        model (Any): Model to be saved.
        path (Union[Path, str]): Path to the directory where to
            save the model.

    Returns:
        InferenceLearner: Model optimized by Speedster.
    """
    if isinstance(model, StableDiffusionPipeline):
        model.unet.model.save(path)
    else:
        model.save(path)



from copy import deepcopy
from typing import Union

from nebullvm.core.models import QuantizationType, DeviceType
from nebullvm.operations.optimizations.compilers.faster_transformer.bert import (  # noqa: E501
    detect_and_swap_bert_model,
)
from nebullvm.operations.optimizations.compilers.torchscript import (
    TorchScriptCompiler,
)
from nebullvm.operations.optimizations.compilers.utils import (
    get_faster_transformer_repo_path,
)
from nebullvm.optional_modules.torch import (
    GraphModule,
    Module,
    ScriptModule,
    torch,
)
from nebullvm.tools.data import DataManager
from nebullvm.tools.huggingface import PyTorchTransformerWrapper

default_lib_path = str(
    get_faster_transformer_repo_path()
    / "build"
    / "lib"
    / "libth_transformer.so"
)


def detect_and_swap_model(model, data_type="fp16", remove_padding=False):
    """currently only supports:
    - BertModel and model with BertModel as .bert attribute
    """
    model = detect_and_swap_bert_model(
        model,
        data_type=data_type,
        lib_path=default_lib_path,
        remove_padding=remove_padding,
    )
    if data_type == "fp16":
        model.half()
    elif data_type == "bf16":
        model.bfloat16()
    return model


class FasterTransformerCompiler(TorchScriptCompiler):
    supported_ops = {
        "cpu": [None, QuantizationType.STATIC, QuantizationType.DYNAMIC],
        "gpu": [
            None,
            QuantizationType.HALF,
        ],
    }

    @torch.no_grad()
    def _compile_model(
        self,
        model: Union[Module, GraphModule],
        input_data: DataManager,
        quantization_type: QuantizationType,
    ) -> ScriptModule:
        model = deepcopy(model)  # Some operations modify the model in-place
        if isinstance(model, PyTorchTransformerWrapper):
            # .core_model is a huggingface model
            data_type = (
                "fp16"
                if quantization_type is QuantizationType.HALF
                else "fp32"
            )
            model.core_model = detect_and_swap_model(
                model.core_model, data_type=data_type, remove_padding=False
            )
            if self.device.type is DeviceType.GPU:
                model.cuda()

        return super()._compile_model(model, input_data, quantization_type)



import logging

from nebullvm.optional_modules.dummy import DummyClass

try:
    import torch_neuron  # noqa F401

    logging.getLogger("Neuron").setLevel(logging.WARNING)
except ImportError:
    try:
        import torch_neuronx  # noqa F401

        logging.getLogger("Neuron").setLevel(logging.WARNING)
    except ImportError:
        torch_neuron = DummyClass



from nebullvm.optional_modules.dummy import DummyClass

try:
    import torch  # noqa F401
    from torch.nn import Module  # noqa F401
    from torch.jit import ScriptModule  # noqa F401
    from torch.fx import GraphModule
    from torch.utils.data import DataLoader, Dataset  # noqa F401
    from torch.quantization.quantize_fx import (  # noqa F401
        prepare_fx,
        convert_fx,
    )

    from torch.ao.quantization.stubs import QuantStub, DeQuantStub
    from torch.fx import symbolic_trace
    from torch.quantization import default_dynamic_qconfig
    import torch.distributed as torch_distributed
except ImportError:

    class nn:
        Module = DummyClass

    class jit:
        ScriptModule = DummyClass

    class fx:
        GraphModule = DummyClass

    class torch:
        float = half = int8 = DummyClass
        float16 = float32 = int32 = int64 = DummyClass
        Tensor = DummyClass
        dtype = DummyClass
        nn = nn
        jit = jit
        Generator = DummyClass
        FloatTensor = DummyClass
        fx = fx

        @staticmethod
        def no_grad():
            return lambda x: None

        @staticmethod
        def inference_mode():
            return lambda x: None

    Dataset = DummyClass
    Module = DummyClass
    ScriptModule = DummyClass
    GraphModule = DummyClass
    DataLoader = DummyClass
    symbolic_trace = None
    QuantStub = DeQuantStub = DummyClass
    default_dynamic_qconfig = prepare_fx = convert_fx = None
    Generator = DummyClass
    FloatTensor = DummyClass
    torch_distributed = None



import abc
import copy
from abc import abstractmethod
import time
from typing import List, Any, Union

from loguru import logger

from nebullvm.core.models import (
    Device,
    DeviceType,
    OptimizedModel,
    OriginalModel,
)
from nebullvm.operations.conversions.huggingface import convert_hf_model
from nebullvm.operations.inference_learners.base import (
    BaseInferenceLearner,
)
from nebullvm.operations.inference_learners.huggingface import (
    DiffusionInferenceLearner,
)
from nebullvm.optional_modules.diffusers import StableDiffusionPipeline
from nebullvm.optional_modules.torch import torch
from nebullvm.tools.diffusers import (
    get_unet_inputs,
    preprocess_diffusers,
    postprocess_diffusers,
)
from nebullvm.tools.pytorch import get_torch_model_size
from nebullvm.tools.utils import (
    is_huggingface_data,
    check_module_version,
    get_throughput,
)


class ModelAdapter(abc.ABC):
    @property
    @abstractmethod
    def adapted_model(self):
        pass

    @property
    @abstractmethod
    def adapted_data(self):
        pass

    @abstractmethod
    def adapt_inference_learner(
        self, optimized_model: OptimizedModel
    ) -> BaseInferenceLearner:
        pass

    @abstractmethod
    def adapt_original_model(
        self, original_model: OriginalModel
    ) -> OriginalModel:
        pass


class DiffusionAdapter(ModelAdapter):
    def __init__(
        self,
        original_pipeline: StableDiffusionPipeline,
        data: List,
        device: Device,
    ):
        self.original_pipeline = copy.deepcopy(original_pipeline)
        self.original_data = data
        self.device = device
        self.__adapted = False
        self.__df_model = None
        self.__df_data = None

    @torch.no_grad()
    def __benchmark_pipeline(
        self,
        pipe: Union[StableDiffusionPipeline, BaseInferenceLearner],
        num_warmup_steps=2,
        num_steps=3,
    ):

        # Warmup
        for i in range(num_warmup_steps):
            _ = pipe(self.original_data[i % len(self.original_data)]).images[0]

        start = time.time()
        # Benchmark
        for i in range(num_steps):
            _ = pipe(self.original_data[i % len(self.original_data)]).images[0]

        took = time.time() - start

        return took / num_steps

    def __adapt(self):
        if not check_module_version(torch, max_version="1.13.1+cu117"):
            raise ValueError(
                "Diffusion models are only supported in PyTorch "
                "versions <= 1.13.1. Please downgrade your PyTorch "
                "version and try again."
            )

        model = copy.deepcopy(self.original_pipeline)
        model.get_unet_inputs = get_unet_inputs
        model.to(self.device.to_torch_format())
        self.__df_data = [
            (
                tuple(
                    d.reshape((1,)) if d.shape == torch.Size([]) else d
                    for d in model.get_unet_inputs(
                        model,
                        prompt=prompt,
                    )
                    if d is not None
                ),
                None,
            )
            for prompt in self.original_data
        ]
        self.__df_model = preprocess_diffusers(model)
        self.__adapted = True

    @property
    def adapted_model(self):
        if self.__adapted is False:
            self.__adapt()
        return self.__df_model

    @property
    def adapted_data(self):
        if self.__adapted is False:
            self.__adapt()
        return self.__df_data

    def adapt_inference_learner(
        self, optimized_model: OptimizedModel
    ) -> OptimizedModel:
        pipe = copy.deepcopy(self.original_pipeline)
        pipe.to(self.device.to_torch_format())
        if self.device.type is DeviceType.GPU:
            try:
                pipe.enable_xformers_memory_efficient_attention()
            except Exception:
                pass

        pipe = postprocess_diffusers(
            optimized_model.inference_learner,
            pipe,
            self.device,
        )
        logger.info("Benchmarking optimized pipeline...")
        optimized_model.latency_seconds = self.__benchmark_pipeline(pipe)
        optimized_model.throughput = get_throughput(
            optimized_model.latency_seconds
        )
        optimized_model.inference_learner = DiffusionInferenceLearner(pipe)
        optimized_model.size_mb += (
            sum(
                [
                    get_torch_model_size(v)
                    for (k, v) in pipe.__dict__.items()
                    if isinstance(v, torch.nn.Module) and k != "unet"
                ]
            )
            / 1e6
        )
        return optimized_model

    def adapt_original_model(
        self, original_model: OriginalModel
    ) -> OriginalModel:
        pipe = copy.deepcopy(self.original_pipeline)
        pipe.to(self.device.to_torch_format())
        logger.info("Benchmarking original pipeline...")
        original_model.latency_seconds = self.__benchmark_pipeline(pipe)
        original_model.throughput = get_throughput(
            original_model.latency_seconds
        )
        original_model.size_mb += (
            sum(
                [
                    get_torch_model_size(v)
                    for (k, v) in pipe.__dict__.items()
                    if isinstance(v, torch.nn.Module) and k != "unet"
                ]
            )
            / 1e6
        )
        return original_model


class HuggingFaceAdapter(ModelAdapter):
    def __init__(self, model: Any, data: List, device: Device, **kwargs):
        self.original_model = model
        self.original_data = data
        self.device = device
        self.tokenizer_params = kwargs
        self.__adapted = False
        self.__hf_model = None
        self.__hf_data = None
        self.__hf_input_names = None
        self.__hf_output_type = None
        self.__hf_output_structure = None

    def __adapt_model(self):
        if not is_huggingface_data(self.original_data[0]):
            raise ValueError("Cannot convert non-HuggingFace data")
        (
            model,
            data,
            input_names,
            output_structure,
            output_type,
        ) = convert_hf_model(
            self.original_model,
            self.original_data,
            self.device,
            **self.tokenizer_params,
        )
        self.__hf_model = model
        self.__hf_data = data
        self.__hf_input_names = input_names
        self.__hf_output_type = output_type
        self.__hf_output_structure = output_structure
        self.__adapted = True

    @property
    def adapted_model(self):
        if self.__adapted is False:
            self.__adapt_model()
        return self.__hf_model

    @property
    def adapted_data(self):
        if self.__adapted is False:
            self.__adapt_model()
        return self.__hf_data

    def adapt_inference_learner(
        self, optimized_model: OptimizedModel
    ) -> OptimizedModel:
        from nebullvm.operations.inference_learners.huggingface import (
            HuggingFaceInferenceLearner,
        )

        optimized_model.inference_learner = HuggingFaceInferenceLearner(
            core_inference_learner=optimized_model.inference_learner,
            output_structure=self.__hf_output_structure,
            input_names=self.__hf_input_names,
            output_type=self.__hf_output_type,
        )

        return optimized_model

    def adapt_original_model(
        self, original_model: OriginalModel
    ) -> OriginalModel:
        return original_model



from nebullvm.core.models import QuantizationType
from nebullvm.operations.optimizations.compilers.torchscript import (
    TorchScriptCompiler,
)
from nebullvm.optional_modules.torch import (
    torch,
)
from nebullvm.tools.data import DataManager


class TorchXLACompiler(TorchScriptCompiler):
    supported_ops = {
        "cpu": [],
        "gpu": [],
        "tpu": [None, QuantizationType.HALF],
    }

    @torch.no_grad()
    def _compile_model(
        self,
        model: torch.nn.Module,
        input_data: DataManager,
        quantization_type: QuantizationType,
    ) -> torch.nn.Module:
        compiled_model = model.to(self.device.to_torch_format())
        return compiled_model



import abc
from typing import Dict, Union

from loguru import logger

from nebullvm.core.models import Device, DeviceType
from nebullvm.tools.feedback_collector import FeedbackCollector
from nebullvm.tools.utils import check_device


class Operation(abc.ABC):
    def __init__(self):
        self._state = {}
        self.device = Device(DeviceType.CPU)
        self.execute_count = 0
        self.logger = logger
        self.feedback_collector = None

    def set_feedback_collector(self, feedback_collector: FeedbackCollector):
        self.feedback_collector = feedback_collector
        for value in self.__dict__.values():
            if isinstance(value, Operation):
                value.set_feedback_collector(feedback_collector)

    @abc.abstractmethod
    def execute(self, **kwargs):
        raise NotImplementedError()

    @property
    def state(self) -> Dict[str, any]:
        return self._state

    def to(self, device: Union[str, Device]):
        if isinstance(device, str):
            self.device = check_device(device)
        else:
            self.device = device
        return self



import unittest
from unittest.mock import patch

from nebullvm.core.models import DeviceType
from nebullvm.tools import utils


class TestGetThroughput(unittest.TestCase):
    def test_latency_is_zero(self):
        self.assertEqual(-1, utils.get_throughput(0, 10))


class TestCheckDevice(unittest.TestCase):
    @patch("nebullvm.tools.utils.gpu_is_available", return_value=False)
    @patch("nebullvm.tools.utils.tpu_is_available", return_value=False)
    @patch("nebullvm.tools.utils.neuron_is_available", return_value=False)
    def test_device_is_none_no_device_available(self, *_):
        device = utils.check_device()
        self.assertEqual(DeviceType.CPU, device.type)
        self.assertEqual(device.idx, 0)

    @patch("nebullvm.tools.utils.gpu_is_available", return_value=True)
    @patch("nebullvm.tools.utils.neuron_is_available", return_value=False)
    @patch("nebullvm.tools.utils.tpu_is_available", return_value=False)
    def test_device_is_none_gpu_is_available(self, *_):
        device = utils.check_device()
        self.assertEqual(DeviceType.GPU, device.type)
        self.assertEqual(device.idx, 0)

    @patch("nebullvm.tools.utils.tpu_is_available", return_value=True)
    @patch("nebullvm.tools.utils.gpu_is_available", return_value=False)
    @patch("nebullvm.tools.utils.neuron_is_available", return_value=False)
    def test_device_is_none_tpu_is_available(self, *_):
        device = utils.check_device()
        self.assertEqual(DeviceType.TPU, device.type)
        self.assertEqual(device.idx, 0)

    @patch("nebullvm.tools.utils.neuron_is_available", return_value=True)
    @patch("nebullvm.tools.utils.gpu_is_available", return_value=False)
    @patch("nebullvm.tools.utils.tpu_is_available", return_value=False)
    def test_device_is_none_neuron_is_available(self, *_):
        device = utils.check_device()
        self.assertEqual(DeviceType.NEURON, device.type)
        self.assertEqual(device.idx, 0)

    def test_device_is_cpu(self):
        device = utils.check_device("cpu")
        self.assertEqual(DeviceType.CPU, device.type)
        self.assertEqual(device.idx, 0)

    @patch("nebullvm.tools.utils.gpu_is_available", return_value=False)
    def test_device_is_gpu_no_gpu_available(self, _):
        device = utils.check_device("gpu")
        self.assertEqual(DeviceType.CPU, device.type)
        self.assertEqual(device.idx, 0)

        device = utils.check_device("cuda")
        self.assertEqual(DeviceType.CPU, device.type)
        self.assertEqual(device.idx, 0)

        device = utils.check_device("cuda:1")
        self.assertEqual(DeviceType.CPU, device.type)
        self.assertEqual(device.idx, 0)

        device = utils.check_device("gpu:2")
        self.assertEqual(DeviceType.CPU, device.type)
        self.assertEqual(device.idx, 0)

    @patch("nebullvm.tools.utils.gpu_is_available", return_value=True)
    def test_device_is_gpu_gpu_is_available(self, _):
        device = utils.check_device("gpu")
        self.assertEqual(DeviceType.GPU, device.type)
        self.assertEqual(device.idx, 0)

        device = utils.check_device("cuda")
        self.assertEqual(DeviceType.GPU, device.type)
        self.assertEqual(device.idx, 0)

        device = utils.check_device("cuda:1")
        self.assertEqual(DeviceType.GPU, device.type)
        self.assertEqual(device.idx, 1)

        device = utils.check_device("gpu:2")
        self.assertEqual(DeviceType.GPU, device.type)
        self.assertEqual(device.idx, 2)

    @patch("nebullvm.tools.utils.tpu_is_available", return_value=False)
    def test_device_is_tpu_no_tpu_available(self, _):
        device = utils.check_device("tpu")
        self.assertEqual(DeviceType.CPU, device.type)
        self.assertEqual(device.idx, 0)

        device = utils.check_device("tpu:1")
        self.assertEqual(DeviceType.CPU, device.type)
        self.assertEqual(device.idx, 0)

    @patch("nebullvm.tools.utils.tpu_is_available", return_value=True)
    def test_device_is_tpu_tpu_is_available(self, _):
        device = utils.check_device("tpu")
        self.assertEqual(DeviceType.TPU, device.type)
        self.assertEqual(device.idx, 0)

        device = utils.check_device("tpu:1")
        self.assertEqual(DeviceType.TPU, device.type)
        self.assertEqual(device.idx, 1)

    @patch("nebullvm.tools.utils.neuron_is_available", return_value=False)
    def test_device_is_neuron_no_neuron_available(self, _):
        device = utils.check_device("neuron")
        self.assertEqual(DeviceType.CPU, device.type)
        self.assertEqual(device.idx, 0)

        device = utils.check_device("neuron:1")
        self.assertEqual(DeviceType.CPU, device.type)
        self.assertEqual(device.idx, 0)

    @patch("nebullvm.tools.utils.neuron_is_available", return_value=True)
    def test_device_is_neuron_neuron_is_available(self, _):
        device = utils.check_device("neuron")
        self.assertEqual(DeviceType.NEURON, device.type)
        self.assertEqual(device.idx, 0)

        device = utils.check_device("neuron:1")
        self.assertEqual(DeviceType.NEURON, device.type)
        self.assertEqual(device.idx, 1)



import unittest
from unittest.mock import patch

from nebullvm.tools import hardware_utils


class TestGetHwSetup(unittest.TestCase):
    @patch(
        "nebullvm.tools.hardware_utils.gpu_is_available", return_value=False
    )
    @patch(
        "nebullvm.tools.hardware_utils.tpu_is_available", return_value=False
    )
    @patch(
        "nebullvm.tools.hardware_utils.neuron_is_available", return_value=False
    )
    def test_hw_setup__gpu_not_available(self, *_):
        setup = hardware_utils.get_hw_setup()
        self.assertIsNone(setup.accelerator)
        self.assertGreater(len(setup.cpu), 0)
        self.assertGreater(len(setup.operating_system), 0)
        self.assertGreater(setup.memory_gb, 0)

    @patch("nebullvm.tools.hardware_utils.gpu_is_available", return_value=True)
    @patch(
        "nebullvm.tools.hardware_utils._get_gpu_name", return_value="mock-gpu"
    )
    def test_hw_setup__gpu_is_available(self, *_):
        setup = hardware_utils.get_hw_setup()
        self.assertEqual("mock-gpu", setup.accelerator)
        self.assertGreater(len(setup.cpu), 0)
        self.assertGreater(len(setup.operating_system), 0)
        self.assertGreater(setup.memory_gb, 0)



from nebullvm.core.models import DeepLearningFramework
from nebullvm.operations.conversions.converters import (
    PytorchConverter,
    TensorflowConverter,
    ONNXConverter,
    Converter,
)


def get_conversion_op(framework: DeepLearningFramework) -> Converter:
    if framework == DeepLearningFramework.PYTORCH:
        conversion_op = PytorchConverter()
    elif framework == DeepLearningFramework.TENSORFLOW:
        conversion_op = TensorflowConverter()
    else:
        conversion_op = ONNXConverter()

    return conversion_op



from nebullvm.installers.auto_installer import (
    select_frameworks_to_install,
    select_compilers_to_install,
)


def test_install_default_option():
    include_frameworks = "all"
    include_backends = "all"

    include_backends = select_frameworks_to_install(
        include_frameworks, include_backends
    )

    assert include_backends == [
        "diffusers",
        "huggingface",
        "onnx",
        "tensorflow",
        "torch",
    ]


def test_install_torch_full():
    include_frameworks = ["torch"]
    include_backends = "all"

    include_backends = select_frameworks_to_install(
        include_frameworks, include_backends
    )

    assert include_backends == ["onnx", "torch"]


def test_install_torch_base():
    include_frameworks = ["torch"]
    include_backends = []

    include_backends = select_frameworks_to_install(
        include_frameworks, include_backends
    )

    assert include_backends == ["torch"]


def test_install_tensorflow_full():
    include_frameworks = ["tensorflow"]
    include_backends = "all"

    include_backends = select_frameworks_to_install(
        include_frameworks, include_backends
    )

    assert include_backends == ["onnx", "tensorflow"]


def test_install_tensorflow_base():
    include_frameworks = ["tensorflow"]
    include_backends = []

    include_backends = select_frameworks_to_install(
        include_frameworks, include_backends
    )

    assert include_backends == ["tensorflow"]


def test_install_onnx_full():
    include_frameworks = ["onnx"]
    include_backends = "all"

    include_backends = select_frameworks_to_install(
        include_frameworks, include_backends
    )

    assert include_backends == ["onnx"]


def test_install_onnx_base():
    include_frameworks = ["onnx"]
    include_backends = []

    include_backends = select_frameworks_to_install(
        include_frameworks, include_backends
    )

    assert include_backends == ["onnx"]


def test_install_diffusers_full():
    include_frameworks = ["diffusers"]
    include_backends = "all"

    include_backends = select_frameworks_to_install(
        include_frameworks, include_backends
    )

    assert include_backends == ["diffusers", "onnx", "torch"]


def test_install_huggingface_full():
    include_frameworks = ["huggingface"]
    include_backends = "all"

    include_backends = select_frameworks_to_install(
        include_frameworks, include_backends
    )

    assert include_backends == ["huggingface", "onnx", "tensorflow", "torch"]


def test_install_huggingface_full_tf():
    include_frameworks = ["huggingface"]
    include_backends = ["onnx", "tensorflow"]

    include_backends = select_frameworks_to_install(
        include_frameworks, include_backends
    )

    assert include_backends == ["huggingface", "onnx", "tensorflow"]


def test_install_huggingface_full_torch():
    include_frameworks = ["huggingface"]
    include_backends = ["onnx", "torch"]

    include_backends = select_frameworks_to_install(
        include_frameworks, include_backends
    )

    assert include_backends == ["huggingface", "onnx", "torch"]


def test_install_huggingface_tf():
    include_frameworks = ["huggingface"]
    include_backends = ["tensorflow"]

    include_backends = select_frameworks_to_install(
        include_frameworks, include_backends
    )

    assert include_backends == ["huggingface", "tensorflow"]


def test_install_huggingface_torch():
    include_frameworks = ["huggingface"]
    include_backends = ["torch"]

    include_backends = select_frameworks_to_install(
        include_frameworks, include_backends
    )

    assert include_backends == ["huggingface", "torch"]


def test_install_huggingface_compilers_all():
    framework_list = ["huggingface"]
    include_compilers = "all"

    compiler_list = select_compilers_to_install(
        include_compilers, framework_list
    )

    assert compiler_list == []


def test_install_huggingface_torch_compilers_all():
    framework_list = ["huggingface", "torch"]
    include_compilers = "all"

    compiler_list = select_compilers_to_install(
        include_compilers, framework_list
    )

    assert compiler_list == [
        "deepsparse",
        "faster_transformer",
        "intel_neural_compressor",
        "tensor_rt",
        "torch_tensor_rt",
    ]


def test_install_torch_compilers_all():
    framework_list = ["torch"]
    include_compilers = "all"

    compiler_list = select_compilers_to_install(
        include_compilers, framework_list
    )

    assert compiler_list == [
        "deepsparse",
        "faster_transformer",
        "intel_neural_compressor",
        "tensor_rt",
        "torch_tensor_rt",
    ]


def test_install_torch_compilers_deepsparse():
    framework_list = ["torch"]
    include_compilers = ["deepsparse"]

    compiler_list = select_compilers_to_install(
        include_compilers, framework_list
    )

    assert compiler_list == ["deepsparse"]


def test_install_torch_compilers_invalid():
    framework_list = ["torch"]
    include_compilers = ["best_compiler"]

    compiler_list = select_compilers_to_install(
        include_compilers, framework_list
    )

    assert compiler_list == []


def test_install_torch_onnx_compilers_all():
    framework_list = ["torch", "onnx"]
    include_compilers = "all"

    compiler_list = select_compilers_to_install(
        include_compilers, framework_list
    )

    assert compiler_list == [
        "deepsparse",
        "faster_transformer",
        "intel_neural_compressor",
        "openvino",
        "tensor_rt",
        "torch_tensor_rt",
    ]


def test_install_tensorflow_compilers_all():
    framework_list = ["tensorflow"]
    include_compilers = "all"

    compiler_list = select_compilers_to_install(
        include_compilers, framework_list
    )

    assert compiler_list == []



import argparse
from typing import List, Union

from loguru import logger

from nebullvm.config import (
    ONNX_MODULES,
    TENSORFLOW_MODULES,
    TORCH_MODULES,
    HUGGING_FACE_MODULES,
    DIFFUSERS_MODULES,
)
from nebullvm.installers.installers import (
    ONNXInstaller,
    PytorchInstaller,
    TensorflowInstaller,
    HuggingFaceInstaller,
    DiffusersInstaller,
)


SUPPORTED_BACKENDS_DICT = {
    "torch": ["onnx"],
    "tensorflow": ["onnx"],
    "huggingface": ["torch", "tensorflow", "onnx"],
    "diffusers": ["torch", "onnx"],
    "onnx": [],
}

INSTALLERS = {
    "onnx": ONNXInstaller,
    "torch": PytorchInstaller,
    "tensorflow": TensorflowInstaller,
    "huggingface": HuggingFaceInstaller,
    "diffusers": DiffusersInstaller,
}

MODULES = {
    "onnx": ONNX_MODULES,
    "torch": TORCH_MODULES,
    "tensorflow": TENSORFLOW_MODULES,
    "huggingface": HUGGING_FACE_MODULES,
    "diffusers": DIFFUSERS_MODULES,
}


def select_frameworks_to_install(
    include_frameworks: Union[List[str], str],
    include_backends: Union[List[str], str],
) -> List[str]:
    supported_frameworks = list(INSTALLERS.keys())
    if isinstance(include_frameworks, str) and include_frameworks == "all":
        frameworks_list = supported_frameworks
    elif isinstance(include_frameworks, list):
        frameworks_list = []
        for framework in include_frameworks:
            if framework in supported_frameworks:
                frameworks_list.append(framework)
            else:
                logger.warning(f"Framework {framework} not supported")

        if len(frameworks_list) == 0:
            raise ValueError("No supported frameworks selected")

        if isinstance(include_backends, str) and include_backends == "all":
            for framework in frameworks_list:
                for backend in SUPPORTED_BACKENDS_DICT[framework]:
                    frameworks_list.append(backend)
        elif isinstance(include_backends, list):
            for backend in include_backends:
                if backend not in supported_frameworks:
                    logger.warning(f"Backend {backend} not supported")
                else:
                    backend_supported = False
                    for framework in frameworks_list:
                        if backend in SUPPORTED_BACKENDS_DICT[framework]:
                            frameworks_list.append(backend)
                            backend_supported = True
                            break
                    if not backend_supported:
                        logger.warning(
                            f"Backend {backend} not supported for selected "
                            f"frameworks"
                        )
        else:
            raise ValueError("Invalid backends list")
    else:
        raise ValueError("Invalid frameworks list")

    frameworks_list = list(set(frameworks_list))
    frameworks_list.sort()

    return frameworks_list


def select_compilers_to_install(
    include_compilers: Union[List[str], str], framework_list: List[str]
) -> List[str]:
    compiler_list = []
    supported_compilers = list(
        set([item for sublist in MODULES.values() for item in sublist])
    )
    if isinstance(include_compilers, str) and include_compilers == "all":
        compiler_list = list(
            set(
                [
                    item
                    for (fr, compilers) in MODULES.items()
                    for item in compilers
                    if fr in framework_list
                ]
            )
        )
    else:
        for compiler in include_compilers:
            if compiler not in supported_compilers:
                logger.warning(f"Compiler {compiler} not supported")
            else:
                compiler_supported = False
                for framework in framework_list:
                    if compiler in MODULES[framework]:
                        compiler_list.append(compiler)
                        compiler_supported = True
                        break
                if not compiler_supported:
                    logger.warning(
                        f"Compiler {compiler} not supported for selected "
                        f"frameworks"
                    )

    compiler_list = list(set(compiler_list))
    compiler_list.sort()

    return compiler_list


def auto_install_libraries(
    include_frameworks: Union[List[str], str] = "all",
    include_backends: Union[List[str], str] = "all",
    include_compilers: Union[List[str], str] = "all",
):
    logger.info("Running auto install of nebullvm dependencies")

    framework_list = select_frameworks_to_install(
        include_frameworks, include_backends
    )

    compilers_list = select_compilers_to_install(
        include_compilers, framework_list
    )

    for framework in framework_list:
        framework_installer = INSTALLERS[framework](MODULES[framework])
        if not framework_installer.check_framework():
            framework_installer.install_framework()
        framework_installer.install_dependencies(framework_list)
        framework_installer.install_compilers(compilers_list)


def main():
    parser = argparse.ArgumentParser(
        description="Auto install dl frameworks and dependencies"
    )
    parser.add_argument(
        "-f",
        "--frameworks",
        help="The base dl frameworks to be installed",
        default="all",
        nargs="+",
    )
    parser.add_argument(
        "-b",
        "--extra-backends",
        help="additional dl frameworks to be installed to "
        "gain the optimal speedup",
        default="all",
        nargs="+",
    )
    parser.add_argument(
        "-c",
        "--compilers",
        help="Compilers to be installed",
        default="all",
        nargs="+",
    )
    args = vars(parser.parse_args())

    if len(args["frameworks"]) == 1 and args["frameworks"][0] == "all":
        framework_list = "all"
    else:
        framework_list = args["frameworks"]

    if len(args["extra_backends"]) == 1 and args["extra_backends"][0] in [
        "all",
        "none",
    ]:
        if args["extra_backends"][0] == "all":
            backend_list = "all"
        else:
            backend_list = []
    else:
        backend_list = args["extra_backends"]

    if len(args["compilers"]) == 1 and args["compilers"][0] == "all":
        compilers_list = "all"
    else:
        compilers_list = args["compilers"]

    auto_install_libraries(framework_list, backend_list, compilers_list)


if __name__ == "__main__":
    main()



# Based on: https://github.com/NVIDIA/FasterTransformer/blob/4402759e48f2340220638675f464b6ba1f79ac3c/examples/pytorch/bert/utils/modeling_bert.py  # noqa: E501
# This file is mostly copied from the FasterTransformer repo
# https://github.com/NVIDIA/FasterTransformer
# Copyright (c) 2020-2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

from loguru import logger

from nebullvm.optional_modules.torch import torch, torch_distributed as dist

from nebullvm.optional_modules.huggingface import (
    BertConfig,
    BertEmbeddings,
    BertEncoder,
    BertPooler,
    BertPreTrainedModel,
)

from .checkpoint_quantization import checkpoint_quantization


class EncoderWeights(object):
    def __init__(
        self,
        layer_num,
        hidden_dim,
        weights=None,
        sparse=False,
        tensor_para_size=1,
        pipeline_para_size=1,
    ):
        """weights need be a state_dict of bert model"""
        self.layer_num = layer_num
        self.int8 = False
        self.hidden_dim = hidden_dim
        self.weights = {}
        self.tensor_para_size = tensor_para_size
        self.pipeline_para_size = pipeline_para_size

        self.use_mpi = dist.is_mpi_available()

        if self.use_mpi:
            try:
                dist.init_process_group(backend="mpi")
            except:  # noqa: E722
                logger.info(
                    "[INFO] WARNING: Exception occurred in "
                    "dist.init_process_group(backend='mpi')."
                    "Maybe the process group has been initialized somewhere else."  # noqa: E501
                )
        else:
            logger.info("[INFO] MPI is not available in this PyTorch build.")
            assert (
                tensor_para_size == 1
            ), "[FATAL] MPI is required for tensor_para_size > 1."
            assert (
                pipeline_para_size == 1
            ), "[FATAL] MPI is required for pipeline_para_size > 1."

        self.rank = dist.get_rank() if self.use_mpi else 0
        self.device_count = torch.cuda.device_count()
        self.device = self.rank % self.device_count
        torch.cuda.set_device(self.device)

        world_size = dist.get_world_size() if self.use_mpi else 1  # noqa: F841
        self.tensor_para_rank = self.rank % self.tensor_para_size
        self.pipeline_para_rank = self.rank // self.tensor_para_size
        if weights is None:
            self._generated_weights = True
            for i in range(layer_num):
                pre = "encoder.layer." + str(i) + "."
                self.weights[
                    pre + "attention.self.query.weight"
                ] = torch.zeros(hidden_dim, hidden_dim)
                self.weights[pre + "attention.self.query.bias"] = torch.zeros(
                    hidden_dim
                )
                self.weights[pre + "attention.self.key.weight"] = torch.zeros(
                    hidden_dim, hidden_dim
                )
                self.weights[pre + "attention.self.key.bias"] = torch.zeros(
                    hidden_dim
                )
                self.weights[
                    pre + "attention.self.value.weight"
                ] = torch.zeros(hidden_dim, hidden_dim)
                self.weights[pre + "attention.self.value.bias"] = torch.zeros(
                    hidden_dim
                )
                self.weights[
                    pre + "attention.output.dense.weight"
                ] = torch.zeros(hidden_dim, hidden_dim)
                self.weights[
                    pre + "attention.output.dense.bias"
                ] = torch.zeros(hidden_dim)
                self.weights[
                    pre + "attention.output.LayerNorm.weight"
                ] = torch.zeros(hidden_dim)
                self.weights[
                    pre + "attention.output.LayerNorm.bias"
                ] = torch.zeros(hidden_dim)
                self.weights[pre + "intermediate.dense.weight"] = torch.zeros(
                    4 * hidden_dim, hidden_dim
                )  # noqa: E501
                self.weights[pre + "intermediate.dense.bias"] = torch.zeros(
                    4 * hidden_dim
                )
                self.weights[pre + "output.dense.weight"] = torch.zeros(
                    hidden_dim, 4 * hidden_dim
                )
                self.weights[pre + "output.dense.bias"] = torch.zeros(
                    hidden_dim
                )
                self.weights[pre + "output.LayerNorm.weight"] = torch.zeros(
                    hidden_dim
                )
                self.weights[pre + "output.LayerNorm.bias"] = torch.zeros(
                    hidden_dim
                )
            for k, v in self.weights.items():
                if not k.endswith("_amax"):
                    self.weights[k] = torch.nn.init.uniform_(v, -1, 1)
            if sparse:
                for k, v in self.weights.items():
                    if (
                        "query.weight" in k
                        or "key.weight" in k
                        or "value.weight" in k
                        or "dense.weight" in k
                    ):
                        v_shape = v.shape
                        v = v.view(-1, 4)
                        _, indices = torch.topk(
                            torch.abs(v), 2, dim=-1, largest=False
                        )
                        v.scatter_(1, indices, 0)
                        self.weights[k] = v.view(v_shape)
        else:
            self._generated_weights = False
            for k, v in weights.items():
                ks = k.split(".")
                if ks[-2] == "LayerNorm":
                    if ks[-1] == "gamma":
                        ks[-1] = "weight"
                    elif ks[-1] == "beta":
                        ks[-1] = "bias"
                self.weights[".".join(ks)] = v

    def listed_weights(self):
        ret = []
        start_layer = (
            self.pipeline_para_rank * self.layer_num // self.pipeline_para_size
        )
        end_layer = (
            (self.pipeline_para_rank + 1)
            * self.layer_num
            // self.pipeline_para_size
        )
        if not self.int8:
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.self.query.weight"
                        ].transpose(-1, -2)
                        for layer_idx in range(start_layer, end_layer)
                    ],
                    0,
                ).contiguous()
            )  # 0
            ret[-1] = (
                ret[-1]
                .split(ret[-1].shape[-1] // self.tensor_para_size, dim=-1)[
                    self.tensor_para_rank
                ]
                .contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.self.query.bias"
                        ]
                        for layer_idx in range(start_layer, end_layer)
                    ],
                    0,
                ).contiguous()
            )
            ret[-1] = (
                ret[-1]
                .split(ret[-1].shape[-1] // self.tensor_para_size, dim=-1)[
                    self.tensor_para_rank
                ]
                .contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.self.key.weight"
                        ].transpose(-1, -2)
                        for layer_idx in range(start_layer, end_layer)
                    ],
                    0,
                ).contiguous()
            )  # 2
            ret[-1] = (
                ret[-1]
                .split(ret[-1].shape[-1] // self.tensor_para_size, dim=-1)[
                    self.tensor_para_rank
                ]
                .contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.self.key.bias"
                        ]
                        for layer_idx in range(start_layer, end_layer)
                    ],
                    0,
                ).contiguous()
            )
            ret[-1] = (
                ret[-1]
                .split(ret[-1].shape[-1] // self.tensor_para_size, dim=-1)[
                    self.tensor_para_rank
                ]
                .contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.self.value.weight"
                        ].transpose(-1, -2)
                        for layer_idx in range(start_layer, end_layer)
                    ],
                    0,
                ).contiguous()
            )  # 4
            ret[-1] = (
                ret[-1]
                .split(ret[-1].shape[-1] // self.tensor_para_size, dim=-1)[
                    self.tensor_para_rank
                ]
                .contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.self.value.bias"
                        ]
                        for layer_idx in range(start_layer, end_layer)
                    ],
                    0,
                ).contiguous()
            )
            ret[-1] = (
                ret[-1]
                .split(ret[-1].shape[-1] // self.tensor_para_size, dim=-1)[
                    self.tensor_para_rank
                ]
                .contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.output.dense.weight"
                        ].transpose(-1, -2)
                        for layer_idx in range(start_layer, end_layer)
                    ],
                    0,
                ).contiguous()
            )  # 6
            ret[-1] = (
                ret[-1]
                .split(ret[-1].shape[1] // self.tensor_para_size, dim=1)[
                    self.tensor_para_rank
                ]
                .contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.output.dense.bias"
                        ]
                        for layer_idx in range(start_layer, end_layer)
                    ],
                    0,
                ).contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.output.LayerNorm.weight"
                        ]
                        for layer_idx in range(start_layer, end_layer)
                    ],
                    0,
                ).contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.output.LayerNorm.bias"
                        ]
                        for layer_idx in range(start_layer, end_layer)
                    ],
                    0,
                ).contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "intermediate.dense.weight"
                        ].transpose(-1, -2)
                        for layer_idx in range(start_layer, end_layer)
                    ],
                    0,
                ).contiguous()
            )  # 10
            ret[-1] = (
                ret[-1]
                .split(ret[-1].shape[-1] // self.tensor_para_size, dim=-1)[
                    self.tensor_para_rank
                ]
                .contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "intermediate.dense.bias"
                        ]
                        for layer_idx in range(start_layer, end_layer)
                    ],
                    0,
                ).contiguous()
            )
            ret[-1] = (
                ret[-1]
                .split(ret[-1].shape[-1] // self.tensor_para_size, dim=-1)[
                    self.tensor_para_rank
                ]
                .contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "output.dense.weight"
                        ].transpose(-1, -2)
                        for layer_idx in range(start_layer, end_layer)
                    ],
                    0,
                ).contiguous()
            )  # 12
            ret[-1] = (
                ret[-1]
                .split(ret[-1].shape[1] // self.tensor_para_size, dim=1)[
                    self.tensor_para_rank
                ]
                .contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "output.dense.bias"
                        ]
                        for layer_idx in range(start_layer, end_layer)
                    ],
                    0,
                ).contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "output.LayerNorm.weight"
                        ]
                        for layer_idx in range(start_layer, end_layer)
                    ],
                    0,
                ).contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "output.LayerNorm.bias"
                        ]
                        for layer_idx in range(start_layer, end_layer)
                    ],
                    0,
                ).contiguous()
            )
        else:
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.self.query.weight"
                        ]
                        for layer_idx in range(self.layer_num)
                    ],
                    0,
                ).contiguous()
            )  # 0
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.self.query.bias"
                        ]
                        for layer_idx in range(self.layer_num)
                    ],
                    0,
                ).contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.self.key.weight"
                        ]
                        for layer_idx in range(self.layer_num)
                    ],
                    0,
                ).contiguous()
            )  # 2
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.self.key.bias"
                        ]
                        for layer_idx in range(self.layer_num)
                    ],
                    0,
                ).contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.self.value.weight"
                        ]
                        for layer_idx in range(self.layer_num)
                    ],
                    0,
                ).contiguous()
            )  # 4
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.self.value.bias"
                        ]
                        for layer_idx in range(self.layer_num)
                    ],
                    0,
                ).contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.output.dense.weight"
                        ]
                        for layer_idx in range(self.layer_num)
                    ],
                    0,
                ).contiguous()
            )  # 6
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.output.dense.bias"
                        ]
                        for layer_idx in range(self.layer_num)
                    ],
                    0,
                ).contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.output.LayerNorm.weight"
                        ]
                        for layer_idx in range(self.layer_num)
                    ],
                    0,
                ).contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "attention.output.LayerNorm.bias"
                        ]
                        for layer_idx in range(self.layer_num)
                    ],
                    0,
                ).contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "intermediate.dense.weight"
                        ]
                        for layer_idx in range(self.layer_num)
                    ],
                    0,
                ).contiguous()
            )  # 10
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "intermediate.dense.bias"
                        ]
                        for layer_idx in range(self.layer_num)
                    ],
                    0,
                ).contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "output.dense.weight"
                        ]
                        for layer_idx in range(self.layer_num)
                    ],
                    0,
                ).contiguous()
            )  # 12
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "output.dense.bias"
                        ]
                        for layer_idx in range(self.layer_num)
                    ],
                    0,
                ).contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "output.LayerNorm.weight"
                        ]
                        for layer_idx in range(self.layer_num)
                    ],
                    0,
                ).contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "output.LayerNorm.bias"
                        ]
                        for layer_idx in range(self.layer_num)
                    ],
                    0,
                ).contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "amaxList"
                        ]
                        for layer_idx in range(self.layer_num)
                    ],
                    0,
                ).contiguous()
            )
            ret.append(
                torch.stack(
                    [
                        self.weights[
                            "encoder.layer."
                            + str(layer_idx)
                            + "."
                            + "h_amaxList"
                        ]
                        for layer_idx in range(self.layer_num)
                    ],
                    0,
                ).contiguous()
            )
        return ret

    def to_cuda(self):
        if not self.int8:
            for k, v in self.weights.items():
                self.weights[k] = v.cuda()
        else:
            h_scale_list = {}
            for k, v in self.weights.items():
                if "amaxList" in k:
                    k_h = k.replace("amaxList", "h_amaxList")
                    h_scale_list[k_h] = v
                self.weights[k] = v.cuda()
            for k, v in h_scale_list.items():
                self.weights[k] = v

    def to_half(self):
        if self.int8:
            raise RuntimeError(
                "Cannot cast to half if the weights have been casted to int8."
            )
        for k, v in self.weights.items():
            self.weights[k] = v.half()

    def to_bfloat16(self):
        if self.int8:
            raise RuntimeError(
                "Cannot cast to bfloat16 if the weights have been casted to int8."  # noqa: E501
            )
        for k, v in self.weights.items():
            self.weights[k] = v.bfloat16()

    def to_int8(self, sparse=False, ths_path="./lib/libth_transformer.so"):
        if self._generated_weights:
            amax_tensor_1 = torch.Tensor(self.hidden_dim).fill_(127.0)
            amax_tensor_2 = torch.Tensor(self.hidden_dim * 4).fill_(127.0)
            for i in range(self.layer_num):
                pre = "encoder.layer." + str(i) + "."
                self.weights[
                    pre + "attention.self.query._input_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "attention.self.query._weight_quantizer._amax"
                ] = amax_tensor_1
                self.weights[
                    pre + "attention.self.query._aftergemm_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "attention.self.key._input_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "attention.self.key._weight_quantizer._amax"
                ] = amax_tensor_1
                self.weights[
                    pre + "attention.self.key._aftergemm_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "attention.self.value._input_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "attention.self.value._weight_quantizer._amax"
                ] = amax_tensor_1
                self.weights[
                    pre + "attention.self.value._aftergemm_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "attention.self.matmul_q_input_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "attention.self.matmul_k_input_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "attention.self.matmul_v_input_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "attention.self.matmul_a_input_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "attention.self.softmax_input_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "attention.output.dense._input_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "attention.output.dense._weight_quantizer._amax"
                ] = amax_tensor_1
                self.weights[
                    pre + "attention.output.dense._aftergemm_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "attention.output.add_local_input_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "attention.output.add_residual_input_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "intermediate.dense._input_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "intermediate.dense._weight_quantizer._amax"
                ] = amax_tensor_2
                self.weights[
                    pre + "intermediate.dense._aftergemm_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "output.dense._input_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "output.dense._weight_quantizer._amax"
                ] = amax_tensor_1
                self.weights[
                    pre + "output.dense._aftergemm_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "output.add_local_input_quantizer._amax"
                ] = torch.tensor(127.0)
                self.weights[
                    pre + "output.add_residual_input_quantizer._amax"
                ] = torch.tensor(127.0)
        if (
            "encoder.layer.0.attention.self.query._input_quantizer._amax"
            not in self.weights
        ):
            raise RuntimeError(
                "There is no quantization node in the checkpoint, cannot be quantized to int8."  # noqa: E501
            )
        if self.int8:
            return
        self.int8 = True
        for k, v in self.weights.items():
            if k.endswith("bias") or k.endswith("LayerNorm.weight"):
                self.weights[k] = v.half()
            elif k.endswith("weight"):
                self.weights[k] = v.float().cuda()
            else:
                self.weights[k] = v.float().cpu()
        self.weights = checkpoint_quantization(
            self.weights, sparse, ths_path, verbose=False
        )


class CustomEncoder(torch.nn.Module):
    def __init__(
        self,
        layer_num,
        head_num,
        head_size,
        weights,
        int8_mode=0,
        remove_padding=False,
        sparse=False,
        path="./lib/libth_transformer.so",
        tensor_para_size=1,
        pipeline_para_size=1,
    ):
        super().__init__()
        self.layer_num = layer_num
        self.remove_padding = remove_padding
        self.int8_mode = int8_mode
        logger.info(f"loading faster transformer library from {path}")
        torch.classes.load_library(path)

        weights_ = weights.listed_weights()

        self.use_mpi = dist.is_mpi_available()

        if self.use_mpi:
            try:
                dist.init_process_group(backend="mpi")
            except:  # noqa: E722
                logger.info(
                    "[INFO] WARNING: Exception occurred in"
                    "dist.init_process_group(backend='mpi')."
                    "Maybe the process group has been initialized somewhere else."  # noqa: E501
                )
        else:
            logger.info("[INFO] MPI is not available in this PyTorch build.")
            assert (
                tensor_para_size == 1
            ), "[FATAL] MPI is required for tensor_para_size > 1."
            assert (
                pipeline_para_size == 1
            ), "[FATAL] MPI is required for pipeline_para_size > 1."

        if int8_mode == 0:
            assert len(weights_) == 16
            try:
                self.encoders = torch.classes.FasterTransformer.Bert(
                    *weights_,
                    head_num,
                    head_size,
                    4 * head_num * head_size,
                    remove_padding,
                    layer_num,
                    sparse,
                    1.0,
                    tensor_para_size,
                    pipeline_para_size,
                )
            except:  # noqa: E722
                # legacy ths for 20.03 image
                self.encoders = torch.classes.FasterTransformerBert(
                    *weights_,
                    head_num,
                    head_size,
                    4 * head_num * head_size,
                    remove_padding,
                    layer_num,
                    sparse,
                    1.0,
                    tensor_para_size,
                    pipeline_para_size,
                )
        else:
            assert len(weights_) == 18
            assert (
                tensor_para_size == 1
            ), "INT8 BERT still only support tensor_para_size = 1"
            assert (
                pipeline_para_size == 1
            ), "INT8 BERT still only support pipeline_para_size = 1"
            try:
                self.encoders = torch.classes.FasterTransformer.INT8Bert(
                    *weights_,
                    head_num,
                    head_size,
                    remove_padding,
                    layer_num,
                    int8_mode,
                    sparse,
                    1.0,
                )
            except:  # noqa: E722
                # legacy ths for 20.03 image
                self.encoders = torch.classes.FasterTransformerINT8Bert(
                    *weights_,
                    head_num,
                    head_size,
                    remove_padding,
                    layer_num,
                    int8_mode,
                    sparse,
                    1.0,
                )

    def forward(self, hidden_states, attention_mask, sequence_lengths):
        hidden_states = self.encoders.forward(hidden_states, sequence_lengths)
        return (hidden_states,)


class HuggingFaceEncoder(torch.nn.Module):
    def __init__(self, layer_num, head_num, head_size, weights=None):
        super().__init__()
        hidden_dim = head_num * head_size
        # TODO(bhsueh) The implementation of hidden_act='gelu' is differen
        #  to FT's (and google BERT) implementation
        # FT's implementation is equivalent to hidden_act='gelu_new',
        # but there are some issues for int8 sparse under gelu_new
        conf = BertConfig(
            hidden_size=hidden_dim,
            intermediate_size=4 * hidden_dim,
            num_attention_heads=head_num,
            num_hidden_layers=layer_num,
            hidden_act="gelu",
        )
        self.encoder = BertEncoder(conf)
        w = {}
        for k, v in weights.weights.items():
            if k.startswith("encoder") and not k.endswith("_amax"):
                w[k[13:]] = weights.weights[k]
        self.encoder.load_state_dict(w)
        self.head_mask = [None] * layer_num

    def forward(self, hidden_states, attention_mask):
        extended_attention_mask = (1.0 - attention_mask) * -10000.0
        output = self.encoder(
            hidden_states,
            extended_attention_mask,
            self.head_mask,
            return_dict=False,
        )
        return output


# Based on: https://github.com/NVIDIA/FasterTransformer/blob/4402759e48f2340220638675f464b6ba1f79ac3c/examples/pytorch/bert/utils/modeling_bert.py # noqa: E501
# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team. # noqa: E501
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PyTorch BERT model modified from HuggingFace transformers. """


class BertModel(BertPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.config = config

        self.embeddings = BertEmbeddings(config)
        self.encoder = BertEncoder(config)
        self.pooler = BertPooler(config)

        self.init_weights()
        self.use_ext_encoder = False

    def forward(
        self,
        input_ids: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.Tensor] = None,
        past_key_values: Optional[List[torch.FloatTensor]] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ):
        if input_ids is not None and inputs_embeds is not None:
            raise ValueError(
                "You cannot specify both input_ids and inputs_embeds at the same time"  # noqa: E501
            )
        elif input_ids is not None:
            input_shape = input_ids.size()
        elif inputs_embeds is not None:
            input_shape = inputs_embeds.size()[:-1]
        else:
            raise ValueError(
                "You have to specify either input_ids or inputs_embeds"
            )

        device = (
            input_ids.device if input_ids is not None else inputs_embeds.device
        )

        if attention_mask is None:
            attention_mask = torch.ones(input_shape, device=device)
        if token_type_ids is None:
            token_type_ids = torch.zeros(
                input_shape, dtype=torch.long, device=device
            )

        if self.use_ext_encoder:
            # if attention_mask.dim() == 3:
            #     extended_attention_mask = attention_mask
            # elif attention_mask.dim() == 2:
            #     extended_attention_mask = attention_mask[:, None, :].repeat(1, input_shape[1], 1) # noqa: E501
            # else:
            #     raise ValueError(
            #         "Wrong shape for input_ids (shape {}) or attention_mask (shape {})".format(# noqa: E501
            #             input_shape, attention_mask.shape
            #         )
            #     )
            assert attention_mask.dim() == 2
            extended_attention_mask = attention_mask.view(
                -1, 1, 1, attention_mask.size(-1)
            )
            m_2 = extended_attention_mask.transpose(-1, -2)
            extended_attention_mask = extended_attention_mask * m_2
            extended_attention_mask = extended_attention_mask.to(
                dtype=next(self.parameters()).dtype
            )  # fp16 compatibility
            seq_lens = torch.sum(attention_mask, 1, dtype=torch.int32).cuda()
        else:
            # We can provide a self-attention mask of dimensions [batch_size, from_seq_length, to_seq_length] # noqa: E501
            # ourselves in which case we just need to make it broadcastable to all heads. # noqa: E501
            if attention_mask.dim() == 3:
                extended_attention_mask = attention_mask[:, None, :, :]
            elif attention_mask.dim() == 2:
                extended_attention_mask = attention_mask[:, None, None, :]
            else:
                raise ValueError(
                    "Wrong shape for input_ids (shape {}) or attention_mask (shape {})".format(  # noqa: E501
                        input_shape, attention_mask.shape
                    )
                )
            # Since attention_mask is 1.0 for positions we want to attend
            # and 0.0 for masked positions, this operation will create a
            # tensor which is 0.0 for positions we want to attend
            # and -10000.0 for masked positions.
            # Since we are adding it to the raw scores before the softmax,
            # this is effectively the same as removing these entirely.
            extended_attention_mask = extended_attention_mask.to(
                dtype=next(self.parameters()).dtype
            )  # fp16 compatibility
            extended_attention_mask = (
                1.0 - extended_attention_mask
            ) * -10000.0

        embedding_output = self.embeddings(
            input_ids=input_ids,
            position_ids=position_ids,
            token_type_ids=token_type_ids,
            inputs_embeds=inputs_embeds,
        )
        if self.use_ext_encoder:
            encoder_outputs = self.encoder(
                embedding_output, extended_attention_mask, seq_lens
            )
        else:
            head_mask = [None] * self.config.num_hidden_layers
            encoder_outputs = self.encoder(
                embedding_output,
                attention_mask=extended_attention_mask,
                head_mask=head_mask,
            )

        sequence_output = encoder_outputs[0]
        pooled_output = self.pooler(sequence_output)

        outputs = (sequence_output, pooled_output,) + encoder_outputs[
            1:
        ]  # add hidden_states and attentions if they are here
        return outputs  # sequence_output, pooled_output, (hidden_states), (attentions) # noqa: E501

    def replace_encoder(self, new_encoder):
        self.encoder = new_encoder
        self.use_ext_encoder = True



import unittest
from unittest.mock import MagicMock

from nebullvm.core.models import OptimizeInferenceResult


class TestOptimizeInferenceResult(unittest.TestCase):
    def test_latency_improvement_rate__optimized_model_is_none(self):
        res = OptimizeInferenceResult(
            original_model=MagicMock(),
            hardware_setup=MagicMock(),
            optimized_model=None,
        )
        self.assertIsNone(res.latency_improvement_rate)

    def test_latency_improvement_rate__optimized_latency_is_zero(self):
        original_latency = 1.0
        optimized_latency = 0.0
        res = OptimizeInferenceResult(
            original_model=MagicMock(latency_seconds=original_latency),
            hardware_setup=MagicMock(),
            optimized_model=MagicMock(latency_seconds=optimized_latency),
        )
        self.assertEqual(-1, res.latency_improvement_rate)

    def test_latency_improvement_rate__original_latency_is_zero(self):
        original_latency = 0.0
        optimized_latency = 1.0
        res = OptimizeInferenceResult(
            original_model=MagicMock(latency_seconds=original_latency),
            hardware_setup=MagicMock(),
            optimized_model=MagicMock(latency_seconds=optimized_latency),
        )
        self.assertEqual(0, res.latency_improvement_rate)

    def test_latency_improvement_rate__rate_gt_1(self):
        original_latency = 1.0
        optimized_latency = 0.5
        res = OptimizeInferenceResult(
            original_model=MagicMock(latency_seconds=original_latency),
            hardware_setup=MagicMock(),
            optimized_model=MagicMock(latency_seconds=optimized_latency),
        )
        self.assertGreater(res.latency_improvement_rate, 1)

    def test_latency_improvement_rate__rate_lt_1(self):
        original_latency = 0.5
        optimized_latency = 1.0
        res = OptimizeInferenceResult(
            original_model=MagicMock(latency_seconds=original_latency),
            hardware_setup=MagicMock(),
            optimized_model=MagicMock(latency_seconds=optimized_latency),
        )
        self.assertLess(res.latency_improvement_rate, 1)

    def test_th_improvement_rate__optimized_model_is_none(self):
        res = OptimizeInferenceResult(
            original_model=MagicMock(),
            hardware_setup=MagicMock(),
            optimized_model=None,
        )
        self.assertIsNone(res.throughput_improvement_rate)

    def test_th_improvement_rate__optimized_th_is_zero(self):
        original_th = 1.0
        optimized_th = 0.0
        res = OptimizeInferenceResult(
            original_model=MagicMock(throughput=original_th),
            hardware_setup=MagicMock(),
            optimized_model=MagicMock(throughput=optimized_th),
        )
        self.assertEqual(0, res.throughput_improvement_rate)

    def test_th_improvement_rate__original_th_is_zero(self):
        original_th = 0.0
        optimized_th = 1.0
        res = OptimizeInferenceResult(
            original_model=MagicMock(throughput=original_th),
            hardware_setup=MagicMock(),
            optimized_model=MagicMock(throughput=optimized_th),
        )
        self.assertEqual(-1, res.throughput_improvement_rate)

    def test_th_improvement_rate__rate_gt_1(self):
        original_th = 0.5
        optimized_th = 1
        res = OptimizeInferenceResult(
            original_model=MagicMock(throughput=original_th),
            hardware_setup=MagicMock(),
            optimized_model=MagicMock(throughput=optimized_th),
        )
        self.assertGreater(res.throughput_improvement_rate, 1)

    def test_th_improvement_rate__rate_lt_1(self):
        original_th = 1.0
        optimized_th = 0.5
        res = OptimizeInferenceResult(
            original_model=MagicMock(throughput=original_th),
            hardware_setup=MagicMock(),
            optimized_model=MagicMock(throughput=optimized_th),
        )
        self.assertLess(res.throughput_improvement_rate, 1)

    def test_size_improvement_rate__optimized_model_is_none(self):
        res = OptimizeInferenceResult(
            original_model=MagicMock(),
            hardware_setup=MagicMock(),
            optimized_model=None,
        )
        self.assertIsNone(res.size_improvement_rate)

    def test_size_improvement_rate__optimized_size_is_zero(self):
        original_size = 1.0
        optimized_size = 0.0
        res = OptimizeInferenceResult(
            original_model=MagicMock(size_mb=original_size),
            hardware_setup=MagicMock(),
            optimized_model=MagicMock(size_mb=optimized_size),
        )
        self.assertEqual(1, res.size_improvement_rate)

    def test_size_improvement_rate__original_size_is_zero(self):
        original_size = 0.0
        optimized_size = 1.0
        res = OptimizeInferenceResult(
            original_model=MagicMock(size_mb=original_size),
            hardware_setup=MagicMock(),
            optimized_model=MagicMock(size_mb=optimized_size),
        )
        self.assertEqual(0, res.size_improvement_rate)

    def test_size_improvement_rate__rate_gt_1(self):
        original_size = 1
        optimized_size = 0.5
        res = OptimizeInferenceResult(
            original_model=MagicMock(size_mb=original_size),
            hardware_setup=MagicMock(),
            optimized_model=MagicMock(size_mb=optimized_size),
        )
        self.assertGreater(res.size_improvement_rate, 1)

    def test_size_improvement_rate__rate_lt_1(self):
        original_size = 0.5
        optimized_size = 1
        res = OptimizeInferenceResult(
            original_model=MagicMock(size_mb=original_size),
            hardware_setup=MagicMock(),
            optimized_model=MagicMock(size_mb=optimized_size),
        )
        self.assertLess(res.size_improvement_rate, 1)

    def test_metric_drop__optimized_model_is_none(self):
        res = OptimizeInferenceResult(
            original_model=MagicMock(),
            hardware_setup=MagicMock(),
            optimized_model=None,
        )
        self.assertIsNone(res.metric_drop)

    def test_metric_drop(self):
        metric_drop = 0.1
        res = OptimizeInferenceResult(
            original_model=MagicMock(),
            hardware_setup=MagicMock(),
            optimized_model=MagicMock(metric_drop=metric_drop),
        )
        self.assertEqual(metric_drop, res.metric_drop)
